from __future__ import annotations

from types import SimpleNamespace

import numpy as np


def sun_position(time: dict | SimpleNamespace, location: dict | SimpleNamespace) -> dict:
    """
    % sun = sun_position(time, location)
    %
    % This function compute the sun position (zenith and azimuth angle at the observer
    % location) as a function of the observer local time and position.
    %
    % It is an implementation of the algorithm presented by Reda et Andreas in:
    %   Reda, I., Andreas, A. (2003) Solar position algorithm for solar
    %   radiation application. National Renewable Energy Laboratory (NREL)
    %   Technical report NREL/TP-560-34302.
    % This document is avalaible at www.osti.gov/bridge
    %
    % This algorithm is based on numerical approximation of the exact equations.
    % The authors of the original paper state that this algorithm should be
    % precise at +/- 0.0003 degrees. I have compared it to NOAA solar table
    % (http://www.srrb.noaa.gov/highlights/sunrise/azel.html) and to USNO solar
    % table (http://aa.usno.navy.mil/data/docs/AltAz.html) and found very good
    % correspondance (up to the precision of those tables), except for large
    % zenith angle, where the refraction by the atmosphere is significant
    % (difference of about 1 degree). Note that in this code the correction
    % for refraction in the atmosphere as been implemented for a temperature
    % of 10C (283 kelvins) and a pressure of 1010 mbar. See the subfunction
    % �sun_topocentric_zenith_angle_calculation� for a possible modification
    % to explicitely model the effect of temperature and pressure as describe
    % in Reda & Andreas (2003).
    %
    % Input parameters:
    %   time: a structure that specify the time when the sun position is
    %   calculated.
    %       time.year: year. Valid for [-2000, 6000]
    %       time.month: month [1-12]
    %       time.day: calendar day [1-31]
    %       time.hour: local hour [0-23]
    %       time.min: minute [0-59]
    %       time.sec: second [0-59]
    %       time.UTC: offset hour from UTC. Local time = Greenwich time + time.UTC
    %   This input can also be passed using the Matlab time format ('dd-mmm-yyyy HH:MM:SS').
    %   In that case, the time has to be specified as UTC time (time.UTC = 0)
    %
    %   location: a structure that specify the location of the observer
    %       location.latitude: latitude (in degrees, north of equator is
    %       positive)
    %       location.longitude: longitude (in degrees, positive for east of
    %       Greenwich)
    %       location.altitude: altitude above mean sea level (in meters)
    %
    % Output parameters
    %   sun: a structure with the calculated sun position
    %       sun.zenith = zenith angle in degrees (angle from the vertical)
    %       sun.azimuth = azimuth angle in degrees, eastward from the north.
    % Only the sun zenith and azimuth angles are returned as output, but a lot
    % of other parameters are calculated that could also extracted as output of
    % this function.
    %
    % Exemple of use
    %
    % location.longitude = -105.1786;
    % location.latitude = 39.742476;
    % location.altitude = 1830.14;
    % time.year = 2005;
    % time.month = 10;
    % time.day = 17;
    % time.hour = 6;
    % time.min = 30;
    % time.sec = 30;
    % time.UTC = -7;
    % %
    % location.longitude = 11.94;
    % location.latitude = 57.70;
    % location.altitude = 3.0;
    % time.UTC = 1;
    % sun = sun_position(time, location);
    %
    % sun =
    %
    %      zenith: 50.1080438859849
    %      azimuth: 194.341174010338
    %
    % History
    %   09/03/2004  Original creation by Vincent Roy (vincent.roy@drdc-rddc.gc.ca)
    %   10/03/2004  Fixed a bug in julian_calculation subfunction (was
    %               incorrect for year 1582 only), Vincent Roy
    %   18/03/2004  Correction to the header (help display) only. No changes to
    %               the code. (changed the �elevation� field in �location� structure
    %               information to �altitude�), Vincent Roy
    %   13/04/2004  Following a suggestion from Jody Klymak (jklymak@ucsd.edu),
    %               allowed the 'time' input to be passed as a Matlab time string.
    %   22/08/2005  Following a bug report from Bruce Bowler
    %               (bbowler@bigelow.org), modified the julian_calculation function. Bug
    %               was 'MATLAB has allowed structure assignment  to a non-empty non-structure
    %               to overwrite the previous value.  This behavior will continue in this release,
    %               but will be an error in a  future version of MATLAB.  For advice on how to
    %               write code that  will both avoid this warning and work in future versions of
    %               MATLAB,  see R14SP2 Release Notes'. Script should now be
    %               compliant with futher release of Matlab...
    """

    # 1. Calculate the Julian Day, and Century. Julian Ephemeris day, century
    # and millenium are calculated using a mean delta_t of 33.184 seconds.
    julian = julian_calculation(time)
    # print(julian)

    # 2. Calculate the Earth heliocentric longitude, latitude, and radius
    # vector (L, B, and R)
    earth_heliocentric_position = earth_heliocentric_position_calculation(julian)

    # 3. Calculate the geocentric longitude and latitude
    sun_geocentric_position = sun_geocentric_position_calculation(earth_heliocentric_position)

    # 4. Calculate the nutation in longitude and obliquity (in degrees).
    nutation = nutation_calculation(julian)

    # 5. Calculate the true obliquity of the ecliptic (in degrees).
    true_obliquity = true_obliquity_calculation(julian, nutation)

    # 6. Calculate the aberration correction (in degrees)
    aberration_correction = abberation_correction_calculation(earth_heliocentric_position)

    # 7. Calculate the apparent sun longitude in degrees)
    apparent_sun_longitude = apparent_sun_longitude_calculation(
        sun_geocentric_position, nutation, aberration_correction
    )

    # 8. Calculate the apparent sideral time at Greenwich (in degrees)
    apparent_stime_at_greenwich = apparent_stime_at_greenwich_calculation(julian, nutation, true_obliquity)

    # 9. Calculate the sun rigth ascension (in degrees)
    sun_rigth_ascension = sun_rigth_ascension_calculation(
        apparent_sun_longitude, true_obliquity, sun_geocentric_position
    )

    # 10. Calculate the geocentric sun declination (in degrees). Positive or
    # negative if the sun is north or south of the celestial equator.
    sun_geocentric_declination = sun_geocentric_declination_calculation(
        apparent_sun_longitude, true_obliquity, sun_geocentric_position
    )

    # 11. Calculate the observer local hour angle (in degrees, westward from south).
    observer_local_hour = observer_local_hour_calculation(apparent_stime_at_greenwich, location, sun_rigth_ascension)

    # 12. Calculate the topocentric sun position (rigth ascension, declination and
    # rigth ascension parallax in degrees)
    topocentric_sun_position = topocentric_sun_position_calculate(
        earth_heliocentric_position, location, observer_local_hour, sun_rigth_ascension, sun_geocentric_declination
    )

    # 13. Calculate the topocentric local hour angle (in degrees)
    topocentric_local_hour = topocentric_local_hour_calculate(observer_local_hour, topocentric_sun_position)

    # 14. Calculate the topocentric zenith and azimuth angle (in degrees)
    sun = sun_topocentric_zenith_angle_calculate(location, topocentric_sun_position, topocentric_local_hour)

    return sun


def julian_calculation(t_input):
    """
    % This function compute the julian day and julian century from the local
    % time and timezone information. Ephemeris are calculated with a delta_t=0
    % seconds.

    % If time input is a Matlab time string, extract the information from
    % this string and create the structure as defined in the main header of
    % this script.
    """
    if not isinstance(t_input, dict):
        # tt = datetime.datetime.strptime(t_input, "%Y-%m-%d %H:%M:%S.%f")    # if t_input is a string of this format
        # t_input should be a datetime object
        time = dict()
        time["UTC"] = 0
        time["year"] = t_input.year
        time["month"] = t_input.month
        time["day"] = t_input.day
        time["hour"] = t_input.hour
        time["min"] = t_input.minute
        time["sec"] = t_input.second
    else:
        time = t_input

    if time["month"] == 1 or time["month"] == 2:
        Y = time["year"] - 1
        M = time["month"] + 12
    else:
        Y = time["year"]
        M = time["month"]

    ut_time = (
        ((time["hour"] - time["UTC"]) / 24) + (time["min"] / (60 * 24)) + (time["sec"] / (60 * 60 * 24))
    )  # time of day in UT time.
    D = time["day"] + ut_time  # Day of month in decimal time, ex. 2sd day of month at 12:30:30UT, D=2.521180556

    # In 1582, the gregorian calendar was adopted
    if time["year"] == 1582:
        if time["month"] == 10:
            if time["day"] <= 4:  # The Julian calendar ended on October 4, 1582
                B = 0
            elif time["day"] >= 15:  # The Gregorian calendar started on October 15, 1582
                A = np.floor(Y / 100)
                B = 2 - A + np.floor(A / 4)
            else:
                print("This date never existed!. Date automatically set to October 4, 1582")
                time["month"] = 10
                time["day"] = 4
                B = 0
        elif time["month"] < 10:  # Julian calendar
            B = 0
        else:  # Gregorian calendar
            A = np.floor(Y / 100)
            B = 2 - A + np.floor(A / 4)
    elif time["year"] < 1582:  # Julian calendar
        B = 0
    else:
        A = np.floor(Y / 100)  # Gregorian calendar
        B = 2 - A + np.floor(A / 4)

    julian = dict()
    julian["day"] = D + B + np.floor(365.25 * (Y + 4716)) + np.floor(30.6001 * (M + 1)) - 1524.5

    delta_t = 0  # 33.184;
    julian["ephemeris_day"] = (julian["day"]) + (delta_t / 86400)
    julian["century"] = (julian["day"] - 2451545) / 36525
    julian["ephemeris_century"] = (julian["ephemeris_day"] - 2451545) / 36525
    julian["ephemeris_millenium"] = (julian["ephemeris_century"]) / 10

    return julian


def earth_heliocentric_position_calculation(julian):
    """
    % This function compute the earth position relative to the sun, using
    % tabulated values.

    % Tabulated values for the longitude calculation
    % L terms  from the original code.
    """
    # Tabulated values for the longitude calculation
    # L terms  from the original code.
    L0_terms = np.array(
        [
            [175347046.0, 0, 0],
            [3341656.0, 4.6692568, 6283.07585],
            [34894.0, 4.6261, 12566.1517],
            [3497.0, 2.7441, 5753.3849],
            [3418.0, 2.8289, 3.5231],
            [3136.0, 3.6277, 77713.7715],
            [2676.0, 4.4181, 7860.4194],
            [2343.0, 6.1352, 3930.2097],
            [1324.0, 0.7425, 11506.7698],
            [1273.0, 2.0371, 529.691],
            [1199.0, 1.1096, 1577.3435],
            [990, 5.233, 5884.927],
            [902, 2.045, 26.298],
            [857, 3.508, 398.149],
            [780, 1.179, 5223.694],
            [753, 2.533, 5507.553],
            [505, 4.583, 18849.228],
            [492, 4.205, 775.523],
            [357, 2.92, 0.067],
            [317, 5.849, 11790.629],
            [284, 1.899, 796.298],
            [271, 0.315, 10977.079],
            [243, 0.345, 5486.778],
            [206, 4.806, 2544.314],
            [205, 1.869, 5573.143],
            [202, 2.4458, 6069.777],
            [156, 0.833, 213.299],
            [132, 3.411, 2942.463],
            [126, 1.083, 20.775],
            [115, 0.645, 0.98],
            [103, 0.636, 4694.003],
            [102, 0.976, 15720.839],
            [102, 4.267, 7.114],
            [99, 6.21, 2146.17],
            [98, 0.68, 155.42],
            [86, 5.98, 161000.69],
            [85, 1.3, 6275.96],
            [85, 3.67, 71430.7],
            [80, 1.81, 17260.15],
            [79, 3.04, 12036.46],
            [71, 1.76, 5088.63],
            [74, 3.5, 3154.69],
            [74, 4.68, 801.82],
            [70, 0.83, 9437.76],
            [62, 3.98, 8827.39],
            [61, 1.82, 7084.9],
            [57, 2.78, 6286.6],
            [56, 4.39, 14143.5],
            [56, 3.47, 6279.55],
            [52, 0.19, 12139.55],
            [52, 1.33, 1748.02],
            [51, 0.28, 5856.48],
            [49, 0.49, 1194.45],
            [41, 5.37, 8429.24],
            [41, 2.4, 19651.05],
            [39, 6.17, 10447.39],
            [37, 6.04, 10213.29],
            [37, 2.57, 1059.38],
            [36, 1.71, 2352.87],
            [36, 1.78, 6812.77],
            [33, 0.59, 17789.85],
            [30, 0.44, 83996.85],
            [30, 2.74, 1349.87],
            [25, 3.16, 4690.48],
        ]
    )

    L1_terms = np.array(
        [
            [628331966747.0, 0, 0],
            [206059.0, 2.678235, 6283.07585],
            [4303.0, 2.6351, 12566.1517],
            [425.0, 1.59, 3.523],
            [119.0, 5.796, 26.298],
            [109.0, 2.966, 1577.344],
            [93, 2.59, 18849.23],
            [72, 1.14, 529.69],
            [68, 1.87, 398.15],
            [67, 4.41, 5507.55],
            [59, 2.89, 5223.69],
            [56, 2.17, 155.42],
            [45, 0.4, 796.3],
            [36, 0.47, 775.52],
            [29, 2.65, 7.11],
            [21, 5.34, 0.98],
            [19, 1.85, 5486.78],
            [19, 4.97, 213.3],
            [17, 2.99, 6275.96],
            [16, 0.03, 2544.31],
            [16, 1.43, 2146.17],
            [15, 1.21, 10977.08],
            [12, 2.83, 1748.02],
            [12, 3.26, 5088.63],
            [12, 5.27, 1194.45],
            [12, 2.08, 4694],
            [11, 0.77, 553.57],
            [10, 1.3, 3286.6],
            [10, 4.24, 1349.87],
            [9, 2.7, 242.73],
            [9, 5.64, 951.72],
            [8, 5.3, 2352.87],
            [6, 2.65, 9437.76],
            [6, 4.67, 4690.48],
        ]
    )

    L2_terms = np.array(
        [
            [52919.0, 0, 0],
            [8720.0, 1.0721, 6283.0758],
            [309.0, 0.867, 12566.152],
            [27, 0.05, 3.52],
            [16, 5.19, 26.3],
            [16, 3.68, 155.42],
            [10, 0.76, 18849.23],
            [9, 2.06, 77713.77],
            [7, 0.83, 775.52],
            [5, 4.66, 1577.34],
            [4, 1.03, 7.11],
            [4, 3.44, 5573.14],
            [3, 5.14, 796.3],
            [3, 6.05, 5507.55],
            [3, 1.19, 242.73],
            [3, 6.12, 529.69],
            [3, 0.31, 398.15],
            [3, 2.28, 553.57],
            [2, 4.38, 5223.69],
            [2, 3.75, 0.98],
        ]
    )

    L3_terms = np.array(
        [
            [289.0, 5.844, 6283.076],
            [35, 0, 0],
            [17, 5.49, 12566.15],
            [3, 5.2, 155.42],
            [1, 4.72, 3.52],
            [1, 5.3, 18849.23],
            [1, 5.97, 242.73],
        ]
    )
    L4_terms = np.array([[114.0, 3.142, 0], [8, 4.13, 6283.08], [1, 3.84, 12566.15]])

    L5_terms = np.array([1, 3.14, 0])
    L5_terms = np.atleast_2d(L5_terms)  # since L5_terms is 1D, we have to convert it to 2D to avoid indexErrors

    A0 = L0_terms[:, 0]
    B0 = L0_terms[:, 1]
    C0 = L0_terms[:, 2]

    A1 = L1_terms[:, 0]
    B1 = L1_terms[:, 1]
    C1 = L1_terms[:, 2]

    A2 = L2_terms[:, 0]
    B2 = L2_terms[:, 1]
    C2 = L2_terms[:, 2]

    A3 = L3_terms[:, 0]
    B3 = L3_terms[:, 1]
    C3 = L3_terms[:, 2]

    A4 = L4_terms[:, 0]
    B4 = L4_terms[:, 1]
    C4 = L4_terms[:, 2]

    A5 = L5_terms[:, 0]
    B5 = L5_terms[:, 1]
    C5 = L5_terms[:, 2]

    JME = julian["ephemeris_millenium"]

    # Compute the Earth Heliochentric longitude from the tabulated values.
    L0 = np.sum(A0 * np.cos(B0 + (C0 * JME)))
    L1 = np.sum(A1 * np.cos(B1 + (C1 * JME)))
    L2 = np.sum(A2 * np.cos(B2 + (C2 * JME)))
    L3 = np.sum(A3 * np.cos(B3 + (C3 * JME)))
    L4 = np.sum(A4 * np.cos(B4 + (C4 * JME)))
    L5 = A5 * np.cos(B5 + (C5 * JME))

    earth_heliocentric_position = dict()
    earth_heliocentric_position["longitude"] = (
        L0
        + (L1 * JME)
        + (L2 * np.power(JME, 2))
        + (L3 * np.power(JME, 3))
        + (L4 * np.power(JME, 4))
        + (L5 * np.power(JME, 5))
    ) / 1e8
    # Convert the longitude to degrees.
    earth_heliocentric_position["longitude"] = earth_heliocentric_position["longitude"] * 180 / np.pi

    # Limit the range to [0,360]
    earth_heliocentric_position["longitude"] = set_to_range(earth_heliocentric_position["longitude"], 0, 360)

    # Tabulated values for the earth heliocentric latitude.
    # B terms  from the original code.
    B0_terms = np.array(
        [[280.0, 3.199, 84334.662], [102.0, 5.422, 5507.553], [80, 3.88, 5223.69], [44, 3.7, 2352.87], [32, 4, 1577.34]]
    )

    B1_terms = np.array([[9, 3.9, 5507.55], [6, 1.73, 5223.69]])

    A0 = B0_terms[:, 0]
    B0 = B0_terms[:, 1]
    C0 = B0_terms[:, 2]

    A1 = B1_terms[:, 0]
    B1 = B1_terms[:, 1]
    C1 = B1_terms[:, 2]

    L0 = np.sum(A0 * np.cos(B0 + (C0 * JME)))
    L1 = np.sum(A1 * np.cos(B1 + (C1 * JME)))

    earth_heliocentric_position["latitude"] = (L0 + (L1 * JME)) / 1e8

    # Convert the latitude to degrees.
    earth_heliocentric_position["latitude"] = earth_heliocentric_position["latitude"] * 180 / np.pi

    # Limit the range to [0,360];
    earth_heliocentric_position["latitude"] = set_to_range(earth_heliocentric_position["latitude"], 0, 360)

    # Tabulated values for radius vector.
    # R terms from the original code
    R0_terms = np.array(
        [
            [100013989.0, 0, 0],
            [1670700.0, 3.0984635, 6283.07585],
            [13956.0, 3.05525, 12566.1517],
            [3084.0, 5.1985, 77713.7715],
            [1628.0, 1.1739, 5753.3849],
            [1576.0, 2.8469, 7860.4194],
            [925.0, 5.453, 11506.77],
            [542.0, 4.564, 3930.21],
            [472.0, 3.661, 5884.927],
            [346.0, 0.964, 5507.553],
            [329.0, 5.9, 5223.694],
            [307.0, 0.299, 5573.143],
            [243.0, 4.273, 11790.629],
            [212.0, 5.847, 1577.344],
            [186.0, 5.022, 10977.079],
            [175.0, 3.012, 18849.228],
            [110.0, 5.055, 5486.778],
            [98, 0.89, 6069.78],
            [86, 5.69, 15720.84],
            [86, 1.27, 161000.69],
            [85, 0.27, 17260.15],
            [63, 0.92, 529.69],
            [57, 2.01, 83996.85],
            [56, 5.24, 71430.7],
            [49, 3.25, 2544.31],
            [47, 2.58, 775.52],
            [45, 5.54, 9437.76],
            [43, 6.01, 6275.96],
            [39, 5.36, 4694],
            [38, 2.39, 8827.39],
            [37, 0.83, 19651.05],
            [37, 4.9, 12139.55],
            [36, 1.67, 12036.46],
            [35, 1.84, 2942.46],
            [33, 0.24, 7084.9],
            [32, 0.18, 5088.63],
            [32, 1.78, 398.15],
            [28, 1.21, 6286.6],
            [28, 1.9, 6279.55],
            [26, 4.59, 10447.39],
        ]
    )

    R1_terms = np.array(
        [
            [103019.0, 1.10749, 6283.07585],
            [1721.0, 1.0644, 12566.1517],
            [702.0, 3.142, 0],
            [32, 1.02, 18849.23],
            [31, 2.84, 5507.55],
            [25, 1.32, 5223.69],
            [18, 1.42, 1577.34],
            [10, 5.91, 10977.08],
            [9, 1.42, 6275.96],
            [9, 0.27, 5486.78],
        ]
    )

    R2_terms = np.array(
        [
            [4359.0, 5.7846, 6283.0758],
            [124.0, 5.579, 12566.152],
            [12, 3.14, 0],
            [9, 3.63, 77713.77],
            [6, 1.87, 5573.14],
            [3, 5.47, 18849],
        ]
    )

    R3_terms = np.array([[145.0, 4.273, 6283.076], [7, 3.92, 12566.15]])

    R4_terms = [4, 2.56, 6283.08]
    R4_terms = np.atleast_2d(R4_terms)  # since L5_terms is 1D, we have to convert it to 2D to avoid indexErrors

    A0 = R0_terms[:, 0]
    B0 = R0_terms[:, 1]
    C0 = R0_terms[:, 2]

    A1 = R1_terms[:, 0]
    B1 = R1_terms[:, 1]
    C1 = R1_terms[:, 2]

    A2 = R2_terms[:, 0]
    B2 = R2_terms[:, 1]
    C2 = R2_terms[:, 2]

    A3 = R3_terms[:, 0]
    B3 = R3_terms[:, 1]
    C3 = R3_terms[:, 2]

    A4 = R4_terms[:, 0]
    B4 = R4_terms[:, 1]
    C4 = R4_terms[:, 2]

    # Compute the Earth heliocentric radius vector
    L0 = np.sum(A0 * np.cos(B0 + (C0 * JME)))
    L1 = np.sum(A1 * np.cos(B1 + (C1 * JME)))
    L2 = np.sum(A2 * np.cos(B2 + (C2 * JME)))
    L3 = np.sum(A3 * np.cos(B3 + (C3 * JME)))
    L4 = A4 * np.cos(B4 + (C4 * JME))

    # Units are in AU
    earth_heliocentric_position["radius"] = (
        L0 + (L1 * JME) + (L2 * np.power(JME, 2)) + (L3 * np.power(JME, 3)) + (L4 * np.power(JME, 4))
    ) / 1e8

    return earth_heliocentric_position


def sun_geocentric_position_calculation(earth_heliocentric_position):
    """
    % This function compute the sun position relative to the earth.
    """
    sun_geocentric_position = dict()
    sun_geocentric_position["longitude"] = earth_heliocentric_position["longitude"] + 180
    # Limit the range to [0,360];
    sun_geocentric_position["longitude"] = set_to_range(sun_geocentric_position["longitude"], 0, 360)

    sun_geocentric_position["latitude"] = -earth_heliocentric_position["latitude"]
    # Limit the range to [0,360]
    sun_geocentric_position["latitude"] = set_to_range(sun_geocentric_position["latitude"], 0, 360)
    return sun_geocentric_position


def nutation_calculation(julian):
    """
    % This function compute the nutation in longtitude and in obliquity, in
    % degrees.
    :param julian:
    :return: nutation
    """

    # All Xi are in degrees.
    JCE = julian["ephemeris_century"]

    # 1. Mean elongation of the moon from the sun
    p = np.atleast_2d([(1 / 189474), -0.0019142, 445267.11148, 297.85036])

    # X0 = polyval(p, JCE);
    X0 = (
        p[0, 0] * np.power(JCE, 3) + p[0, 1] * np.power(JCE, 2) + p[0, 2] * JCE + p[0, 3]
    )  # This is faster than polyval...

    # 2. Mean anomaly of the sun (earth)
    p = np.atleast_2d([-(1 / 300000), -0.0001603, 35999.05034, 357.52772])

    # X1 = polyval(p, JCE)
    X1 = p[0, 0] * np.power(JCE, 3) + p[0, 1] * np.power(JCE, 2) + p[0, 2] * JCE + p[0, 3]

    # 3. Mean anomaly of the moon
    p = np.atleast_2d([(1 / 56250), 0.0086972, 477198.867398, 134.96298])

    # X2 = polyval(p, JCE);
    X2 = p[0, 0] * np.power(JCE, 3) + p[0, 1] * np.power(JCE, 2) + p[0, 2] * JCE + p[0, 3]

    # 4. Moon argument of latitude
    p = np.atleast_2d([(1 / 327270), -0.0036825, 483202.017538, 93.27191])

    # X3 = polyval(p, JCE)
    X3 = p[0, 0] * np.power(JCE, 3) + p[0, 1] * np.power(JCE, 2) + p[0, 2] * JCE + p[0, 3]

    # 5. Longitude of the ascending node of the moon's mean orbit on the
    # ecliptic, measured from the mean equinox of the date
    p = np.atleast_2d([(1 / 450000), 0.0020708, -1934.136261, 125.04452])

    # X4 = polyval(p, JCE);
    X4 = p[0, 0] * np.power(JCE, 3) + p[0, 1] * np.power(JCE, 2) + p[0, 2] * JCE + p[0, 3]

    # Y tabulated terms from the original code
    Y_terms = np.array(
        [
            [0, 0, 0, 0, 1],
            [-2, 0, 0, 2, 2],
            [0, 0, 0, 2, 2],
            [0, 0, 0, 0, 2],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [-2, 1, 0, 2, 2],
            [0, 0, 0, 2, 1],
            [0, 0, 1, 2, 2],
            [-2, -1, 0, 2, 2],
            [-2, 0, 1, 0, 0],
            [-2, 0, 0, 2, 1],
            [0, 0, -1, 2, 2],
            [2, 0, 0, 0, 0],
            [0, 0, 1, 0, 1],
            [2, 0, -1, 2, 2],
            [0, 0, -1, 0, 1],
            [0, 0, 1, 2, 1],
            [-2, 0, 2, 0, 0],
            [0, 0, -2, 2, 1],
            [2, 0, 0, 2, 2],
            [0, 0, 2, 2, 2],
            [0, 0, 2, 0, 0],
            [-2, 0, 1, 2, 2],
            [0, 0, 0, 2, 0],
            [-2, 0, 0, 2, 0],
            [0, 0, -1, 2, 1],
            [0, 2, 0, 0, 0],
            [2, 0, -1, 0, 1],
            [-2, 2, 0, 2, 2],
            [0, 1, 0, 0, 1],
            [-2, 0, 1, 0, 1],
            [0, -1, 0, 0, 1],
            [0, 0, 2, -2, 0],
            [2, 0, -1, 2, 1],
            [2, 0, 1, 2, 2],
            [0, 1, 0, 2, 2],
            [-2, 1, 1, 0, 0],
            [0, -1, 0, 2, 2],
            [2, 0, 0, 2, 1],
            [2, 0, 1, 0, 0],
            [-2, 0, 2, 2, 2],
            [-2, 0, 1, 2, 1],
            [2, 0, -2, 0, 1],
            [2, 0, 0, 0, 1],
            [0, -1, 1, 0, 0],
            [-2, -1, 0, 2, 1],
            [-2, 0, 0, 0, 1],
            [0, 0, 2, 2, 1],
            [-2, 0, 2, 0, 1],
            [-2, 1, 0, 2, 1],
            [0, 0, 1, -2, 0],
            [-1, 0, 1, 0, 0],
            [-2, 1, 0, 0, 0],
            [1, 0, 0, 0, 0],
            [0, 0, 1, 2, 0],
            [0, 0, -2, 2, 2],
            [-1, -1, 1, 0, 0],
            [0, 1, 1, 0, 0],
            [0, -1, 1, 2, 2],
            [2, -1, -1, 2, 2],
            [0, 0, 3, 2, 2],
            [2, -1, 0, 2, 2],
        ]
    )

    nutation_terms = np.array(
        [
            [-171996, -174.2, 92025, 8.9],
            [-13187, -1.6, 5736, -3.1],
            [-2274, -0.2, 977, -0.5],
            [2062, 0.2, -895, 0.5],
            [1426, -3.4, 54, -0.1],
            [712, 0.1, -7, 0],
            [-517, 1.2, 224, -0.6],
            [-386, -0.4, 200, 0],
            [-301, 0, 129, -0.1],
            [217, -0.5, -95, 0.3],
            [-158, 0, 0, 0],
            [129, 0.1, -70, 0],
            [123, 0, -53, 0],
            [63, 0, 0, 0],
            [63, 0.1, -33, 0],
            [-59, 0, 26, 0],
            [-58, -0.1, 32, 0],
            [-51, 0, 27, 0],
            [48, 0, 0, 0],
            [46, 0, -24, 0],
            [-38, 0, 16, 0],
            [-31, 0, 13, 0],
            [29, 0, 0, 0],
            [29, 0, -12, 0],
            [26, 0, 0, 0],
            [-22, 0, 0, 0],
            [21, 0, -10, 0],
            [17, -0.1, 0, 0],
            [16, 0, -8, 0],
            [-16, 0.1, 7, 0],
            [-15, 0, 9, 0],
            [-13, 0, 7, 0],
            [-12, 0, 6, 0],
            [11, 0, 0, 0],
            [-10, 0, 5, 0],
            [-8, 0, 3, 0],
            [7, 0, -3, 0],
            [-7, 0, 0, 0],
            [-7, 0, 3, 0],
            [-7, 0, 3, 0],
            [6, 0, 0, 0],
            [6, 0, -3, 0],
            [6, 0, -3, 0],
            [-6, 0, 3, 0],
            [-6, 0, 3, 0],
            [5, 0, 0, 0],
            [-5, 0, 3, 0],
            [-5, 0, 3, 0],
            [-5, 0, 3, 0],
            [4, 0, 0, 0],
            [4, 0, 0, 0],
            [4, 0, 0, 0],
            [-4, 0, 0, 0],
            [-4, 0, 0, 0],
            [-4, 0, 0, 0],
            [3, 0, 0, 0],
            [-3, 0, 0, 0],
            [-3, 0, 0, 0],
            [-3, 0, 0, 0],
            [-3, 0, 0, 0],
            [-3, 0, 0, 0],
            [-3, 0, 0, 0],
            [-3, 0, 0, 0],
        ]
    )

    # Using the tabulated values, compute the delta_longitude and
    # delta_obliquity.
    Xi = np.array([X0, X1, X2, X3, X4])  # a col mat in octave

    tabulated_argument = Y_terms.dot(np.transpose(Xi)) * (np.pi / 180)

    delta_longitude = (nutation_terms[:, 0] + (nutation_terms[:, 1] * JCE)) * np.sin(tabulated_argument)
    delta_obliquity = (nutation_terms[:, 2] + (nutation_terms[:, 3] * JCE)) * np.cos(tabulated_argument)

    nutation = dict()  # init nutation dictionary
    # Nutation in longitude
    nutation["longitude"] = np.sum(delta_longitude) / 36000000

    # Nutation in obliquity
    nutation["obliquity"] = np.sum(delta_obliquity) / 36000000

    return nutation


def true_obliquity_calculation(julian, nutation):
    """
    This function compute the true obliquity of the ecliptic.

    :param julian:
    :param nutation:
    :return:
    """

    p = np.atleast_2d([2.45, 5.79, 27.87, 7.12, -39.05, -249.67, -51.38, 1999.25, -1.55, -4680.93, 84381.448])

    # mean_obliquity = polyval(p, julian.ephemeris_millenium/10);
    U = julian["ephemeris_millenium"] / 10
    mean_obliquity = (
        p[0, 0] * np.power(U, 10)
        + p[0, 1] * np.power(U, 9)
        + p[0, 2] * np.power(U, 8)
        + p[0, 3] * np.power(U, 7)
        + p[0, 4] * np.power(U, 6)
        + p[0, 5] * np.power(U, 5)
        + p[0, 6] * np.power(U, 4)
        + p[0, 7] * np.power(U, 3)
        + p[0, 8] * np.power(U, 2)
        + p[0, 9] * U
        + p[0, 10]
    )

    true_obliquity = (mean_obliquity / 3600) + nutation["obliquity"]

    return true_obliquity


def abberation_correction_calculation(earth_heliocentric_position):
    """
    This function compute the aberration_correction, as a function of the
    earth-sun distance.

    :param earth_heliocentric_position:
    :return:
    """
    aberration_correction = -20.4898 / (3600 * earth_heliocentric_position["radius"])
    return aberration_correction


def apparent_sun_longitude_calculation(sun_geocentric_position, nutation, aberration_correction):
    """
    This function compute the sun apparent longitude

    :param sun_geocentric_position:
    :param nutation:
    :param aberration_correction:
    :return:
    """
    apparent_sun_longitude = sun_geocentric_position["longitude"] + nutation["longitude"] + aberration_correction
    return apparent_sun_longitude


def apparent_stime_at_greenwich_calculation(julian, nutation, true_obliquity):
    """
    This function compute the apparent sideral time at Greenwich.

    :param julian:
    :param nutation:
    :param true_obliquity:
    :return:
    """

    JD = julian["day"]
    JC = julian["century"]

    # Mean sideral time, in degrees
    mean_stime = (
        280.46061837
        + (360.98564736629 * (JD - 2451545))
        + (0.000387933 * np.power(JC, 2))
        - (np.power(JC, 3) / 38710000)
    )

    # Limit the range to [0-360];
    mean_stime = set_to_range(mean_stime, 0, 360)

    apparent_stime_at_greenwich = mean_stime + (nutation["longitude"] * np.cos(true_obliquity * np.pi / 180))
    return apparent_stime_at_greenwich


def sun_rigth_ascension_calculation(apparent_sun_longitude, true_obliquity, sun_geocentric_position):
    """
    This function compute the sun rigth ascension.
    :param apparent_sun_longitude:
    :param true_obliquity:
    :param sun_geocentric_position:
    :return:
    """

    argument_numerator = (np.sin(apparent_sun_longitude * np.pi / 180) * np.cos(true_obliquity * np.pi / 180)) - (
        np.tan(sun_geocentric_position["latitude"] * np.pi / 180) * np.sin(true_obliquity * np.pi / 180)
    )
    argument_denominator = np.cos(apparent_sun_longitude * np.pi / 180)
    sun_rigth_ascension = np.arctan2(argument_numerator, argument_denominator) * 180 / np.pi
    # Limit the range to [0,360];
    sun_rigth_ascension = set_to_range(sun_rigth_ascension, 0, 360)
    return sun_rigth_ascension


def sun_geocentric_declination_calculation(apparent_sun_longitude, true_obliquity, sun_geocentric_position):
    """

    :param apparent_sun_longitude:
    :param true_obliquity:
    :param sun_geocentric_position:
    :return:
    """

    argument = (np.sin(sun_geocentric_position["latitude"] * np.pi / 180) * np.cos(true_obliquity * np.pi / 180)) + (
        np.cos(sun_geocentric_position["latitude"] * np.pi / 180)
        * np.sin(true_obliquity * np.pi / 180)
        * np.sin(apparent_sun_longitude * np.pi / 180)
    )

    sun_geocentric_declination = np.arcsin(argument) * 180 / np.pi
    return sun_geocentric_declination


def observer_local_hour_calculation(apparent_stime_at_greenwich, location, sun_rigth_ascension):
    """
    This function computes observer local hour.

    :param apparent_stime_at_greenwich:
    :param location:
    :param sun_rigth_ascension:
    :return:
    """

    observer_local_hour = apparent_stime_at_greenwich + location["longitude"] - sun_rigth_ascension
    # Set the range to [0-360]
    observer_local_hour = set_to_range(observer_local_hour, 0, 360)
    return observer_local_hour


def topocentric_sun_position_calculate(
    earth_heliocentric_position, location, observer_local_hour, sun_rigth_ascension, sun_geocentric_declination
):
    """
    This function compute the sun position (rigth ascension and declination)
    with respect to the observer local position at the Earth surface.

    :param earth_heliocentric_position:
    :param location:
    :param observer_local_hour:
    :param sun_rigth_ascension:
    :param sun_geocentric_declination:
    :return:
    """

    # Equatorial horizontal parallax of the sun in degrees
    eq_horizontal_parallax = 8.794 / (3600 * earth_heliocentric_position["radius"])

    # Term u, used in the following calculations (in radians)
    u = np.arctan(0.99664719 * np.tan(location["latitude"] * np.pi / 180))

    # Term x, used in the following calculations
    x = np.cos(u) + ((location["altitude"] / 6378140) * np.cos(location["latitude"] * np.pi / 180))

    # Term y, used in the following calculations
    y = (0.99664719 * np.sin(u)) + ((location["altitude"] / 6378140) * np.sin(location["latitude"] * np.pi / 180))

    # Parallax in the sun rigth ascension (in radians)
    nominator = -x * np.sin(eq_horizontal_parallax * np.pi / 180) * np.sin(observer_local_hour * np.pi / 180)
    denominator = np.cos(sun_geocentric_declination * np.pi / 180) - (
        x * np.sin(eq_horizontal_parallax * np.pi / 180) * np.cos(observer_local_hour * np.pi / 180)
    )
    sun_rigth_ascension_parallax = np.arctan2(nominator, denominator)
    # Conversion to degrees.
    topocentric_sun_position = dict()
    topocentric_sun_position["rigth_ascension_parallax"] = sun_rigth_ascension_parallax * 180 / np.pi

    # Topocentric sun rigth ascension (in degrees)
    topocentric_sun_position["rigth_ascension"] = sun_rigth_ascension + (sun_rigth_ascension_parallax * 180 / np.pi)

    # Topocentric sun declination (in degrees)
    nominator = (
        np.sin(sun_geocentric_declination * np.pi / 180) - (y * np.sin(eq_horizontal_parallax * np.pi / 180))
    ) * np.cos(sun_rigth_ascension_parallax)
    denominator = np.cos(sun_geocentric_declination * np.pi / 180) - (
        y * np.sin(eq_horizontal_parallax * np.pi / 180)
    ) * np.cos(observer_local_hour * np.pi / 180)
    topocentric_sun_position["declination"] = np.arctan2(nominator, denominator) * 180 / np.pi
    return topocentric_sun_position


def topocentric_local_hour_calculate(observer_local_hour, topocentric_sun_position):
    """
    This function compute the topocentric local jour angle in degrees

    :param observer_local_hour:
    :param topocentric_sun_position:
    :return:
    """

    topocentric_local_hour = observer_local_hour - topocentric_sun_position["rigth_ascension_parallax"]
    return topocentric_local_hour


def sun_topocentric_zenith_angle_calculate(location, topocentric_sun_position, topocentric_local_hour):
    """
    This function compute the sun zenith angle, taking into account the
    atmospheric refraction. A default temperature of 283K and a
    default pressure of 1010 mbar are used.

    :param location:
    :param topocentric_sun_position:
    :param topocentric_local_hour:
    :return:
    """

    # Topocentric elevation, without atmospheric refraction
    argument = (
        np.sin(location["latitude"] * np.pi / 180) * np.sin(topocentric_sun_position["declination"] * np.pi / 180)
    ) + (
        np.cos(location["latitude"] * np.pi / 180)
        * np.cos(topocentric_sun_position["declination"] * np.pi / 180)
        * np.cos(topocentric_local_hour * np.pi / 180)
    )
    true_elevation = np.arcsin(argument) * 180 / np.pi

    # Atmospheric refraction correction (in degrees)
    argument = true_elevation + (10.3 / (true_elevation + 5.11))
    refraction_corr = 1.02 / (60 * np.tan(argument * np.pi / 180))

    # For exact pressure and temperature correction, use this,
    # with P the pressure in mbar amd T the temperature in Kelvins:
    # refraction_corr = (P/1010) * (283/T) * 1.02 / (60 * tan(argument * pi/180));

    # Apparent elevation
    apparent_elevation = true_elevation + refraction_corr

    sun = dict()
    sun["zenith"] = 90 - apparent_elevation

    # Topocentric azimuth angle. The +180 conversion is to pass from astronomer
    # notation (westward from south) to navigation notation (eastward from
    # north);
    nominator = np.sin(topocentric_local_hour * np.pi / 180)
    denominator = (np.cos(topocentric_local_hour * np.pi / 180) * np.sin(location["latitude"] * np.pi / 180)) - (
        np.tan(topocentric_sun_position["declination"] * np.pi / 180) * np.cos(location["latitude"] * np.pi / 180)
    )
    sun["azimuth"] = (np.arctan2(nominator, denominator) * 180 / np.pi) + 180

    # Set the range to [0-360]
    sun["azimuth"] = set_to_range(sun["azimuth"], 0, 360)
    return sun


def set_to_range(var, min_interval, max_interval):
    """
    Sets a variable in range min_interval and max_interval

    :param var:
    :param min_interval:
    :param max_interval:
    :return:
    """
    var = var - max_interval * np.floor(var / max_interval)

    if var < min_interval:
        var = var + max_interval
    return var



"""Model configuration and human-body parameter classes.

Defines :class:`ModelConfig` (run-time settings such as sky model,
tiling, and shadow distance) and :class:`HumanParams` (posture,
absorption coefficients, and PET body parameters).
"""

from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    pass


@dataclass
class ModelConfig:
    """
    Model configuration for SOLWEIG calculations.

    Groups all computational settings in one typed object.
    Pure configuration - no paths or data.

    Attributes:
        use_anisotropic_sky: Use Perez anisotropic sky model for diffuse
            radiation. Default True. Requires precomputed shadow matrices.
        human: Human body parameters for Tmrt calculations.
        outputs: Which outputs to save in timeseries calculations.
            Default ``["tmrt"]``.
        physics: Physics parameters (vegetation, posture geometry). Optional.
        materials: Material properties (albedo, emissivity). Optional.
        max_shadow_distance_m: Maximum shadow reach in metres. Default 1000.0.
            Caps horizontal shadow ray distance and serves as tile overlap
            buffer for automatic tiled processing. On mountainous terrain,
            increase this to capture terrain shadows across valleys.
        tile_workers: Number of workers for tiled orchestration. If None,
            picks an adaptive default based on CPU count.
        tile_queue_depth: Extra queued tile tasks beyond active workers.
            If None, defaults to one queue slot per worker when prefetching
            is enabled.
        prefetch_tiles: Whether to prefetch tile tasks beyond active workers.
            If None, runtime chooses automatically based on memory pressure.

    Note:
        UTCI and PET are computed inline when requested via
        ``outputs=["utci"]``.

    Examples:
        Basic usage with defaults:

        >>> config = ModelConfig.defaults()
        >>> config.save("my_config.json")

        Custom configuration:

        >>> config = ModelConfig(
        ...     use_anisotropic_sky=True,
        ...     human=HumanParams(abs_k=0.7, posture="standing"),
        ... )

        Load from legacy JSON:

        >>> config = ModelConfig.from_json("parametersforsolweig.json")
    """

    use_anisotropic_sky: bool = True
    human: HumanParams | None = None
    outputs: list[str] = field(default_factory=lambda: ["tmrt"])
    physics: SimpleNamespace | None = None
    materials: SimpleNamespace | None = None
    max_shadow_distance_m: float = 1000.0
    tile_workers: int | None = None
    tile_queue_depth: int | None = None
    prefetch_tiles: bool | None = None

    def __post_init__(self):
        """Validate configuration fields.

        Note: ``human`` is intentionally left as None here.
        Default ``HumanParams()`` is instantiated by ``calculate()``
        when no human params are provided.
        """
        if self.max_shadow_distance_m <= 0:
            raise ValueError(f"max_shadow_distance_m must be > 0, got {self.max_shadow_distance_m}")
        if self.tile_workers is not None and self.tile_workers < 1:
            raise ValueError(f"tile_workers must be >= 1, got {self.tile_workers}")
        if self.tile_queue_depth is not None and self.tile_queue_depth < 0:
            raise ValueError(f"tile_queue_depth must be >= 0, got {self.tile_queue_depth}")

    @classmethod
    def defaults(cls) -> ModelConfig:
        """
        Create a ModelConfig with recommended defaults.

        Returns:
            ModelConfig with ``use_anisotropic_sky=True`` and all other
            fields at their dataclass defaults.
        """
        return cls(
            use_anisotropic_sky=True,
        )

    @classmethod
    def from_json(cls, path: str | Path) -> ModelConfig:
        """
        Load configuration from legacy JSON parameters file.

        Args:
            path: Path to parametersforsolweig.json

        Returns:
            ModelConfig with settings extracted from JSON

        Example:
            >>> config = ModelConfig.from_json("parametersforsolweig.json")
            >>> config.human.abs_k  # From Tmrt_params
            0.7
        """
        from ..loaders import load_params

        params = load_params(path)

        # Extract human parameters from JSON
        human = HumanParams()
        if hasattr(params, "Tmrt_params"):
            human.abs_k = getattr(params.Tmrt_params, "absK", 0.7)
            human.abs_l = getattr(params.Tmrt_params, "absL", 0.97)
            posture_str = getattr(params.Tmrt_params, "posture", "Standing")
            human.posture = posture_str.lower()

        if hasattr(params, "PET_settings"):
            human.age = getattr(params.PET_settings, "Age", 35)
            human.weight = getattr(params.PET_settings, "Weight", 75.0)
            human.height = getattr(params.PET_settings, "Height", 1.75)
            human.sex = getattr(params.PET_settings, "Sex", 1)
            human.activity = getattr(params.PET_settings, "Activity", 80.0)
            human.clothing = getattr(params.PET_settings, "clo", 0.9)

        return cls(
            human=human,
            materials=params,
        )

    def save(self, path: str | Path):
        """
        Save configuration to JSON file.

        Args:
            path: Output path for JSON file

        Example:
            >>> config = ModelConfig.defaults()
            >>> config.save("my_settings.json")
        """
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        # Serialize to dict
        data = {
            "use_anisotropic_sky": self.use_anisotropic_sky,
            "max_shadow_distance_m": self.max_shadow_distance_m,
            "tile_workers": self.tile_workers,
            "tile_queue_depth": self.tile_queue_depth,
            "prefetch_tiles": self.prefetch_tiles,
            "outputs": self.outputs,
            "human": {
                "posture": self.human.posture,
                "abs_k": self.human.abs_k,
                "abs_l": self.human.abs_l,
                "age": self.human.age,
                "weight": self.human.weight,
                "height": self.human.height,
                "sex": self.human.sex,
                "activity": self.human.activity,
                "clothing": self.human.clothing,
            }
            if self.human
            else None,
        }

        with open(path, "w") as f:
            json.dump(data, f, indent=2)

        logger.info(f"Saved configuration to {path}")

    @classmethod
    def load(cls, path: str | Path) -> ModelConfig:
        """
        Load configuration from JSON file.

        Args:
            path: Path to JSON configuration file

        Returns:
            ModelConfig loaded from file

        Example:
            >>> config = ModelConfig.load("my_settings.json")
            >>> results = calculate(surface, weather, config=config)
        """
        path = Path(path)

        with open(path) as f:
            data = json.load(f)

        # Deserialize human params
        human = None
        if data.get("human"):
            human = HumanParams(**data["human"])

        return cls(
            use_anisotropic_sky=data.get("use_anisotropic_sky", True),
            max_shadow_distance_m=data.get("max_shadow_distance_m", 1000.0),
            tile_workers=data.get("tile_workers"),
            tile_queue_depth=data.get("tile_queue_depth"),
            prefetch_tiles=data.get("prefetch_tiles"),
            human=human,
            outputs=data.get("outputs", ["tmrt"]),
        )


@dataclass
class HumanParams:
    """
    Human body parameters for thermal comfort calculations.

    These parameters affect how radiation is absorbed by a person.
    Default values represent a standard reference person.

    Attributes:
        posture: Body posture ("standing" or "sitting"). Default "standing".
        abs_k: Shortwave absorption coefficient. Default 0.7.
        abs_l: Longwave absorption coefficient. Default 0.97.

    PET-specific parameters (used when ``"pet"`` is in outputs):
        age: Age in years. Default 35.
        weight: Body weight in kg. Default 75.
        height: Body height in meters. Default 1.75.
        sex: Biological sex (1=male, 2=female). Default 1.
        activity: Metabolic activity in W. Default 80.
        clothing: Clothing insulation in clo. Default 0.9.
    """

    posture: str = "standing"
    abs_k: float = 0.7
    abs_l: float = 0.97

    # PET-specific (optional)
    age: int = 35
    weight: float = 75.0
    height: float = 1.75
    sex: int = 1
    activity: float = 80.0
    clothing: float = 0.9

    def __post_init__(self):
        valid_postures = ("standing", "sitting")
        if self.posture not in valid_postures:
            raise ValueError(f"Posture must be one of {valid_postures}, got {self.posture}")
        if not 0 < self.abs_k <= 1:
            raise ValueError(f"abs_k must be in (0, 1], got {self.abs_k}")
        if not 0 < self.abs_l <= 1:
            raise ValueError(f"abs_l must be in (0, 1], got {self.abs_l}")
        if not 0 <= self.age <= 150:
            raise ValueError(f"age must be in [0, 150], got {self.age}")
        if self.weight <= 0:
            raise ValueError(f"weight must be > 0, got {self.weight}")
        if not 0.5 <= self.height <= 2.5:
            raise ValueError(f"height must be in [0.5, 2.5] m, got {self.height}")
        if self.sex not in (1, 2):
            raise ValueError(f"sex must be 1 (male) or 2 (female), got {self.sex}")
        if self.activity < 0:
            raise ValueError(f"activity must be >= 0, got {self.activity}")
        if self.clothing < 0:
            raise ValueError(f"clothing must be >= 0, got {self.clothing}")



"""
Tiled processing for large rasters.

This module provides automatic and manual tiling for SOLWEIG calculations,
supporting both single-timestep and timeseries modes. Large rasters are
automatically divided into overlapping tiles with buffers sized to capture
shadows from the tallest buildings at low sun angles.

Timeseries mode preserves thermal state accumulation across tiles and timesteps,
ensuring physically accurate ground temperature modeling with thermal inertia.
"""

from __future__ import annotations

import math
import os
import sys
import time
from collections.abc import Callable
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any

import numpy as np

from .errors import MissingPrecomputedData
from .models import HumanParams, PrecomputedData, SolweigResult, SurfaceData, ThermalState, TileSpec
from .output_async import AsyncGeoTiffWriter, async_output_enabled
from .postprocess import compute_utci_grid
from .solweig_logging import get_logger
from .summary import GridAccumulator, TimeseriesSummary

logger = get_logger(__name__)

_TIMING_ENABLED = os.environ.get("SOLWEIG_TIMING", "").lower() in ("1", "true")

if TYPE_CHECKING:
    from .models import (
        Location,
        ModelConfig,
        Weather,
    )


# =============================================================================
# Constants
# =============================================================================

MIN_TILE_SIZE = 256  # Minimum tile size in pixels
_FALLBACK_MAX_TILE_SIZE = 2500  # Used when GPU + RAM detection both fail
MIN_SUN_ELEVATION_DEG = 3.0  # Minimum sun elevation for shadow calculations
MAX_BUFFER_M = 1000.0  # Default maximum buffer / shadow distance in meters

# Backward-compat alias (imported by tests)
MAX_TILE_SIZE = _FALLBACK_MAX_TILE_SIZE

# Resource estimation constants
_RAM_FRACTION = 0.50  # Use at most 50% of total physical RAM for tile arrays

# GPU memory constants — two sets per context, derived from
# rust/src/gpu/shadow_gpu.rs allocate_buffers() + init_svf_accumulation().
#
# *_TOTAL: aggregate footprint of ALL wgpu buffers alive simultaneously.
#          Used when max_buffer_size ≈ total GPU memory (Metal/DX12).
# *_SINGLE: largest single buffer per pixel.
#           Used when max_buffer_size is a per-buffer cap (Vulkan/GL).
#
# Shadow context ("solweig" timestep): allocate_buffers() only
#   16 f32 storage buffers (64 B/px) + staging 10× (40 B/px) + overhead/headroom
_SHADOW_GPU_TOTAL_BPP = 120
_SHADOW_GPU_SINGLE_BPP = 40  # staging_buffer = 10 × buffer_size
#
# SVF context: allocate_buffers() + init_svf_accumulation()
#   shadow (104) + svf_data 15× (60) + svf_staging (60)
#   + bitpack 3×20 output (60) + bitpack staging (60)
_SVF_GPU_TOTAL_BPP = 384
_SVF_GPU_SINGLE_BPP = 60  # svf_data_buffer = 15 × buffer_size
#
# SVF CPU/RAM: Rust SvfIntermediate::zeros() = 15 f32 arrays (60 B/px)
# + 3 bitpacked u8 arrays at 20 B each (60 B/px) + memmap overhead (~30 B/px).
_SVF_RAM_BYTES_PER_PIXEL = 150

_SOLWEIG_BYTES_PER_PIXEL = 400  # Peak Python-side bytes per pixel (benchmarked ~370)
_GPU_HEADROOM = 0.80  # Use 80% of GPU max buffer to leave headroom

# Backends where max_buffer_size ≈ total GPU/unified memory.
# On these backends we constrain by total allocation across all buffers.
# On others (Vulkan, GL) max_buffer_size is a per-buffer cap, so we
# constrain by the largest single buffer and rely on init_svf_accumulation()
# falling back to CPU if total VRAM is exceeded.
_TOTAL_MEMORY_BACKENDS = {"Metal", "Dx12"}
_MAX_AUTO_TILE_WORKERS = min(os.cpu_count() or 4, 16)  # Scale with CPU count, cap at 16

# Cache for computed tile limits (populated once per context on first call)
_cached_max_tile_side: dict[str, int] = {}


# =============================================================================
# Resource detection
# =============================================================================


def _get_total_ram_bytes() -> int | None:
    """
    Detect total physical RAM in bytes.

    Uses ``os.sysconf`` on POSIX (macOS/Linux) and ``ctypes`` on Windows.
    Returns ``None`` if detection fails.  No external dependencies.
    """
    import os
    import sys

    try:
        if sys.platform == "win32":
            import ctypes

            class MEMORYSTATUSEX(ctypes.Structure):
                _fields_ = [
                    ("dwLength", ctypes.c_ulong),
                    ("dwMemoryLoad", ctypes.c_ulong),
                    ("ullTotalPhys", ctypes.c_ulonglong),
                    ("ullAvailPhys", ctypes.c_ulonglong),
                    ("ullTotalPageFile", ctypes.c_ulonglong),
                    ("ullAvailPageFile", ctypes.c_ulonglong),
                    ("ullTotalVirtual", ctypes.c_ulonglong),
                    ("ullAvailVirtual", ctypes.c_ulonglong),
                    ("ullAvailExtendedVirtual", ctypes.c_ulonglong),
                ]

            stat = MEMORYSTATUSEX()
            stat.dwLength = ctypes.sizeof(stat)
            ctypes.windll.kernel32.GlobalMemoryStatusEx(ctypes.byref(stat))
            return stat.ullTotalPhys
        else:
            pages = os.sysconf("SC_PHYS_PAGES")
            page_size = os.sysconf("SC_PAGE_SIZE")
            if pages > 0 and page_size > 0:
                return pages * page_size
    except (OSError, ValueError, AttributeError):
        pass
    return None


def compute_max_tile_pixels(*, context: str = "solweig") -> int:
    """
    Compute the maximum number of pixels that fit in a single tile,
    based on real GPU buffer limits and system RAM.

    Args:
        context: ``"solweig"`` for timestep tiling, or ``"svf"`` for SVF-only
            tiling. Affects the bytes-per-pixel estimate used for the RAM
            constraint.

    Returns:
        Maximum pixel count for a tile (rows * cols).
    """
    from . import get_gpu_limits

    if context == "svf":
        ram_bytes_per_pixel = _SVF_RAM_BYTES_PER_PIXEL
        gpu_total_bpp = _SVF_GPU_TOTAL_BPP
        gpu_single_bpp = _SVF_GPU_SINGLE_BPP
    else:
        ram_bytes_per_pixel = _SOLWEIG_BYTES_PER_PIXEL
        gpu_total_bpp = _SHADOW_GPU_TOTAL_BPP
        gpu_single_bpp = _SHADOW_GPU_SINGLE_BPP

    # GPU constraint: total GPU allocation for all buffers must fit in memory.
    # On Metal (Apple Silicon), max_buffer_size ≈ total GPU/unified memory,
    # so this effectively constrains the aggregate footprint per tile.
    # - "solweig": shadow buffers only (~104 B/px)
    # - "svf": shadow + SVF accumulation + bitpack (~344 B/px with veg)
    gpu_max_pixels = None
    limits = get_gpu_limits()
    if limits is not None:
        max_buf = limits["max_buffer_size"]
        backend = str(limits.get("backend", ""))
        # Metal/DX12 backends generally report max_buffer_size close to total
        # GPU/unified memory, so constrain by aggregate per-tile working set.
        # Other backends may expose per-buffer caps; for them use largest single
        # buffer estimate.
        bpp = gpu_total_bpp if backend in _TOTAL_MEMORY_BACKENDS else gpu_single_bpp
        gpu_max_pixels = int(int(max_buf) * _GPU_HEADROOM) // bpp

    # RAM constraint: total physical RAM × fraction / bytes per pixel.
    # - "svf": ~150 B/px (Rust SvfIntermediate + memmap overhead)
    # - "solweig": ~400 B/px (timestep with radiation grids, state arrays, etc).
    ram_max_pixels = None
    total_ram = _get_total_ram_bytes()
    if total_ram is not None:
        usable_ram = int(total_ram * _RAM_FRACTION)
        ram_max_pixels = usable_ram // ram_bytes_per_pixel

    # Use the tighter of the two constraints
    candidates = [c for c in [gpu_max_pixels, ram_max_pixels] if c is not None]
    if candidates:
        return max(MIN_TILE_SIZE**2, min(candidates))

    # Fallback: no detection succeeded
    return _FALLBACK_MAX_TILE_SIZE**2


def compute_max_tile_side(*, context: str = "solweig") -> int:
    """
    Compute the maximum tile side length (square tiles) from resource limits.

    The result is cached per *context* for the lifetime of the process.

    Returns:
        Maximum tile side in pixels (at least ``MIN_TILE_SIZE``).
    """
    import math

    if context in _cached_max_tile_side:
        return _cached_max_tile_side[context]

    max_pixels = compute_max_tile_pixels(context=context)
    side = max(MIN_TILE_SIZE, int(math.isqrt(max_pixels)))

    # Log once so the user can see what limits are driving tile sizing
    from . import get_gpu_limits

    limits = get_gpu_limits()
    total_ram = _get_total_ram_bytes()
    gpu_str = f"{limits['max_buffer_size']:,} bytes" if limits else "N/A"
    ram_str = f"{total_ram:,} bytes" if total_ram else "N/A"
    logger.info(
        f"Resource-aware tile sizing (context={context}): "
        f"GPU max_buffer={gpu_str}, system RAM={ram_str}, max_tile_side={side} px"
    )

    _cached_max_tile_side[context] = side
    return side


# =============================================================================
# Helper Functions
# =============================================================================


def _should_use_tiling(rows: int, cols: int) -> bool:
    """Check if raster size exceeds resource limits and requires tiling."""
    max_side = compute_max_tile_side(context="solweig")
    return rows > max_side or cols > max_side


def _calculate_auto_tile_size(rows: int, cols: int) -> int:
    """
    Calculate optimal core tile size based on raster dimensions and resources.

    Returns the resource-derived maximum tile side as the core size.
    ``validate_tile_size()`` will further adjust to ensure the full tile
    (core + 2 × overlap) fits within resource limits.

    Returns:
        Core tile size in pixels.
    """
    return compute_max_tile_side(context="solweig")


def _resolve_tile_workers(tile_workers: int | None, n_tiles: int) -> int:
    """Resolve worker count for tiled orchestration."""
    if n_tiles <= 0:
        return 1
    if tile_workers is not None and tile_workers < 1:
        raise ValueError(f"tile_workers must be >= 1, got {tile_workers}")
    if tile_workers is None:
        cpu_count = os.cpu_count() or 2
        tile_workers = max(2, min(_MAX_AUTO_TILE_WORKERS, cpu_count // 2))
    return max(1, min(tile_workers, n_tiles))


def _resolve_inflight_limit(
    n_workers: int,
    n_tiles: int,
    tile_queue_depth: int | None,
    prefetch_tiles: bool,
) -> int:
    """
    Resolve max number of in-flight tile tasks.

    ``tile_queue_depth`` controls queued tasks beyond active workers.
    Effective in-flight task limit is ``n_workers + queue_depth``.
    """
    if tile_queue_depth is not None and tile_queue_depth < 0:
        raise ValueError(f"tile_queue_depth must be >= 0, got {tile_queue_depth}")

    queue_depth = (n_workers if prefetch_tiles else 0) if tile_queue_depth is None else tile_queue_depth

    return max(1, min(n_tiles, n_workers + queue_depth))


def _resolve_prefetch_default(
    n_workers: int,
    n_tiles: int,
    core_tile_size: int,
    buffer_pixels: int,
) -> bool:
    """
    Decide default prefetch behavior based on estimated in-flight memory pressure.

    Uses a conservative estimate of per-tile Python-side working memory.
    Prefetch is enabled only when estimated in-flight bytes are comfortably
    below the usable RAM budget.
    """
    if n_tiles <= 0:
        return False

    total_ram = _get_total_ram_bytes()
    if total_ram is None:
        return True

    full_side = core_tile_size + 2 * buffer_pixels
    tile_pixels = max(MIN_TILE_SIZE**2, full_side * full_side)
    estimated_tile_bytes = tile_pixels * _SOLWEIG_BYTES_PER_PIXEL

    # Default prefetch queues up to n_workers extra tasks.
    estimated_inflight_tiles = min(n_tiles, n_workers * 2)
    estimated_inflight_bytes = estimated_inflight_tiles * estimated_tile_bytes

    usable_ram = int(total_ram * _RAM_FRACTION)
    return estimated_inflight_bytes <= int(usable_ram * 0.5)


def _maybe_subdivide_single_tile_for_timeseries(
    rows: int,
    cols: int,
    tile_size: int,
    buffer_pixels: int,
    pixel_size: float,
    requested_workers: int | None,
) -> int:
    """
    Optionally reduce tile size for large single-tile timeseries runs.

    Motivation:
    Resource-aware sizing often yields one very large tile that fits memory, but
    that can leave CPU/GPU orchestration under-utilized in timeseries mode.
    Splitting into a few tiles enables overlapping GPU and CPU work across workers.
    """
    if rows * cols < 4_000_000:
        return tile_size

    env_target = os.getenv("SOLWEIG_TIMESERIES_TARGET_TILES", "").strip()
    try:
        target_tiles = int(env_target) if env_target else 0
    except ValueError:
        target_tiles = 0

    if target_tiles <= 1:
        if requested_workers is not None:
            target_tiles = max(2, min(16, requested_workers))
        else:
            cpu_count = os.cpu_count() or 2
            target_tiles = max(2, min(16, cpu_count // 2))

    splits = max(2, int(math.ceil(math.sqrt(target_tiles))))
    candidate_core = int(math.ceil(max(rows / splits, cols / splits)))
    candidate_core = max(MIN_TILE_SIZE, min(candidate_core, tile_size))

    if candidate_core >= tile_size:
        return tile_size

    adjusted_candidate, warning = validate_tile_size(candidate_core, buffer_pixels, pixel_size)
    if warning:
        logger.warning(warning)
    if adjusted_candidate >= tile_size:
        return tile_size

    candidate_tiles = generate_tiles(rows, cols, adjusted_candidate, buffer_pixels)
    if len(candidate_tiles) <= 1:
        return tile_size

    logger.info(
        "Timeseries CPU parallelization: splitting single tile into "
        f"{len(candidate_tiles)} tiles (core {tile_size} -> {adjusted_candidate})"
    )
    return adjusted_candidate


def _extract_tile_surface(
    surface: SurfaceData,
    tile: TileSpec,
    pixel_size: float,
    precomputed: PrecomputedData | None = None,
) -> SurfaceData:
    """
    Extract tile slice from full surface, reusing precomputed SVF when available.

    Creates a new SurfaceData with sliced arrays (DSM, CDSM, etc.).
    If the global surface has precomputed SVF (via prepare() or compute_svf()),
    the SVF is sliced to the tile bounds — avoiding expensive per-tile
    recomputation. If surface.svf is absent but precomputed.svf is provided,
    that precomputed SVF is sliced instead. When neither source exists, SVF
    remains unset and callers must fail fast before computation.

    Args:
        surface: Full raster surface data.
        tile: Tile specification with slice bounds.
        pixel_size: Pixel size in meters.
        precomputed: Optional precomputed data containing SVF.

    Returns:
        SurfaceData for this tile.
    """
    read_slice = tile.read_slice

    # Use views (no .copy()) for read-only surface arrays. These are never mutated
    # during computation — Rust receives PyReadonlyArray2 and Python never writes.
    # Views avoid redundant allocation, especially for overlapping tile buffers.
    tile_dsm = surface.dsm[read_slice]
    tile_cdsm = surface.cdsm[read_slice] if surface.cdsm is not None else None
    tile_tdsm = surface.tdsm[read_slice] if surface.tdsm is not None else None
    tile_dem = surface.dem[read_slice] if surface.dem is not None else None
    tile_lc = surface.land_cover[read_slice] if surface.land_cover is not None else None
    tile_albedo = surface.albedo[read_slice] if surface.albedo is not None else None
    tile_emis = surface.emissivity[read_slice] if surface.emissivity is not None else None
    tile_wall_ht = surface.wall_height[read_slice] if surface.wall_height is not None else None
    tile_wall_asp = surface.wall_aspect[read_slice] if surface.wall_aspect is not None else None

    # Slice precomputed SVF if available (avoids per-tile recomputation)
    tile_svf = None
    if surface.svf is not None:
        tile_svf = surface.svf.crop(
            tile.row_start_full,
            tile.row_end_full,
            tile.col_start_full,
            tile.col_end_full,
        )
    elif precomputed is not None and precomputed.svf is not None:
        tile_svf = precomputed.svf.crop(
            tile.row_start_full,
            tile.row_end_full,
            tile.col_start_full,
            tile.col_end_full,
        )

    # Slice shadow matrices if available (required for anisotropic sky in tiled mode)
    tile_shadow_matrices = None
    if surface.shadow_matrices is not None:
        tile_shadow_matrices = surface.shadow_matrices.crop(
            tile.row_start_full,
            tile.row_end_full,
            tile.col_start_full,
            tile.col_end_full,
        )

    tile_surface = SurfaceData(
        dsm=tile_dsm,
        cdsm=tile_cdsm,
        tdsm=tile_tdsm,
        dem=tile_dem,
        land_cover=tile_lc,
        albedo=tile_albedo,
        emissivity=tile_emis,
        wall_height=tile_wall_ht,
        wall_aspect=tile_wall_asp,
        pixel_size=pixel_size,
        svf=tile_svf,
        shadow_matrices=tile_shadow_matrices,
    )
    return tile_surface


def _slice_tile_precomputed(
    precomputed: PrecomputedData | None,
    tile: TileSpec,
) -> PrecomputedData | None:
    """
    Slice walls and shadow matrices from precomputed data for a tile.

    SVF is handled via surface.svf (sliced in _extract_tile_surface).
    Shadow matrices are spatially cropped to the tile bounds for
    anisotropic sky support in tiled mode.

    Args:
        precomputed: Full raster precomputed data (or None).
        tile: Tile specification with slice bounds.

    Returns:
        PrecomputedData with sliced walls and shadow matrices, or None.
    """
    if precomputed is None:
        return None

    read_slice = tile.read_slice

    tile_wall_ht = None
    tile_wall_asp = None
    tile_shadow_matrices = None

    if precomputed.wall_height is not None:
        tile_wall_ht = precomputed.wall_height[read_slice].copy()
    if precomputed.wall_aspect is not None:
        tile_wall_asp = precomputed.wall_aspect[read_slice].copy()
    if precomputed.shadow_matrices is not None:
        tile_shadow_matrices = precomputed.shadow_matrices.crop(
            tile.row_start_full,
            tile.row_end_full,
            tile.col_start_full,
            tile.col_end_full,
        )

    if tile_wall_ht is None and tile_wall_asp is None and tile_shadow_matrices is None:
        return None

    return PrecomputedData(
        wall_height=tile_wall_ht,
        wall_aspect=tile_wall_asp,
        svf=None,
        shadow_matrices=tile_shadow_matrices,
    )


def _write_tile_result(
    tile_result: SolweigResult,
    tile: TileSpec,
    tmrt_out: np.ndarray,
    shadow_out: np.ndarray | None,
    kdown_out: np.ndarray | None,
    kup_out: np.ndarray | None,
    ldown_out: np.ndarray | None,
    lup_out: np.ndarray | None,
) -> None:
    """Write core region of tile result to global output arrays."""
    core_slice = tile.core_slice
    write_slice = tile.write_slice

    tmrt_out[write_slice] = tile_result.tmrt[core_slice]
    if shadow_out is not None and tile_result.shadow is not None:
        shadow_out[write_slice] = tile_result.shadow[core_slice]
    if kdown_out is not None and tile_result.kdown is not None:
        kdown_out[write_slice] = tile_result.kdown[core_slice]
    if kup_out is not None and tile_result.kup is not None:
        kup_out[write_slice] = tile_result.kup[core_slice]
    if ldown_out is not None and tile_result.ldown is not None:
        ldown_out[write_slice] = tile_result.ldown[core_slice]
    if lup_out is not None and tile_result.lup is not None:
        lup_out[write_slice] = tile_result.lup[core_slice]


def _slice_tile_state(state: ThermalState, tile: TileSpec) -> ThermalState:
    """
    Slice thermal state arrays for a tile.

    Spatial arrays are sliced using tile.read_slice (full tile with overlap).
    Scalar values are copied as-is (they're global, not spatial).

    Args:
        state: Global thermal state for full raster.
        tile: Tile specification with slice bounds.

    Returns:
        ThermalState for this tile.
    """
    read_slice = tile.read_slice

    return ThermalState(
        tgmap1=state.tgmap1[read_slice].copy(),
        tgmap1_e=state.tgmap1_e[read_slice].copy(),
        tgmap1_s=state.tgmap1_s[read_slice].copy(),
        tgmap1_w=state.tgmap1_w[read_slice].copy(),
        tgmap1_n=state.tgmap1_n[read_slice].copy(),
        tgout1=state.tgout1[read_slice].copy(),
        firstdaytime=state.firstdaytime,
        timeadd=state.timeadd,
        timestep_dec=state.timestep_dec,
    )


def _refresh_tile_state(tile_state: ThermalState, global_state: ThermalState, tile: TileSpec) -> None:
    """
    Refresh a preallocated tile state from the global state in-place.

    This avoids reallocating ThermalState objects/arrays each timestep while
    still ensuring overlap regions are synchronized from the latest global state.
    """
    read_slice = tile.read_slice
    np.copyto(tile_state.tgmap1, global_state.tgmap1[read_slice])
    np.copyto(tile_state.tgmap1_e, global_state.tgmap1_e[read_slice])
    np.copyto(tile_state.tgmap1_s, global_state.tgmap1_s[read_slice])
    np.copyto(tile_state.tgmap1_w, global_state.tgmap1_w[read_slice])
    np.copyto(tile_state.tgmap1_n, global_state.tgmap1_n[read_slice])
    np.copyto(tile_state.tgout1, global_state.tgout1[read_slice])
    tile_state.firstdaytime = global_state.firstdaytime
    tile_state.timeadd = global_state.timeadd
    tile_state.timestep_dec = global_state.timestep_dec


def _merge_tile_state(
    tile_state: ThermalState,
    tile: TileSpec,
    global_state: ThermalState,
) -> None:
    """
    Merge tile state arrays back into global state (in-place).

    Writes core region (tile.core_slice) of tile state arrays to the
    corresponding region (tile.write_slice) in global state. Updates
    global scalar values from tile state (identical across all tiles
    for a given timestep).

    Args:
        tile_state: Computed state for this tile.
        tile: Tile specification with slice bounds.
        global_state: Global state to update (modified in-place).
    """
    core_slice = tile.core_slice
    write_slice = tile.write_slice

    global_state.tgmap1[write_slice] = tile_state.tgmap1[core_slice]
    global_state.tgmap1_e[write_slice] = tile_state.tgmap1_e[core_slice]
    global_state.tgmap1_s[write_slice] = tile_state.tgmap1_s[core_slice]
    global_state.tgmap1_w[write_slice] = tile_state.tgmap1_w[core_slice]
    global_state.tgmap1_n[write_slice] = tile_state.tgmap1_n[core_slice]
    global_state.tgout1[write_slice] = tile_state.tgout1[core_slice]

    # Scalars are the same across all tiles for a given timestep
    global_state.firstdaytime = tile_state.firstdaytime
    global_state.timeadd = tile_state.timeadd


# =============================================================================
# Public Functions
# =============================================================================


def calculate_buffer_distance(
    max_height: float,
    min_sun_elev_deg: float = MIN_SUN_ELEVATION_DEG,
    max_shadow_distance_m: float = MAX_BUFFER_M,
) -> float:
    """
    Calculate required buffer distance for tiled processing based on max building height.

    The buffer must be large enough to capture shadows cast by the tallest buildings
    at the lowest sun elevation angle.

    Formula: buffer = min(max_height / tan(min_sun_elevation), max_shadow_distance_m)

    Args:
        max_height: Maximum building/DSM height in meters.
        min_sun_elev_deg: Minimum sun elevation angle in degrees. Default 3.0.
        max_shadow_distance_m: Maximum buffer distance in meters. Default 1000.0.

    Returns:
        Buffer distance in meters, capped at max_shadow_distance_m.

    Example:
        >>> calculate_buffer_distance(30.0)  # 30m building
        572.9  # 30m / tan(3°)
        >>> calculate_buffer_distance(10.0)  # 10m building
        190.8  # 10m / tan(3°)
    """
    if max_height <= 0:
        return 0.0

    tan_elev = np.tan(np.radians(min_sun_elev_deg))
    if tan_elev <= 0:
        return max_shadow_distance_m

    buffer = max_height / tan_elev
    return min(buffer, max_shadow_distance_m)


def validate_tile_size(
    tile_size: int,
    buffer_pixels: int,
    pixel_size: float,
    context: str = "solweig",
) -> tuple[int, str | None]:
    """
    Validate and adjust core tile size for tiled processing.

    ``tile_size`` is the **core** tile side (the region whose results are
    kept). The actual tile in memory is ``core + 2 × buffer_pixels``.
    This function ensures the full tile fits within resource-derived limits.

    Args:
        tile_size: Requested core tile size in pixels.
        buffer_pixels: Overlap buffer size in pixels.
        pixel_size: Pixel size in meters.
        context: Resource context for limit detection. Use ``"svf"`` for
            SVF preprocessing tiles and ``"solweig"`` for timestep tiles.

    Returns:
        Tuple of (adjusted_core_size, warning_message or None).

    Constraints:
        - core >= MIN_TILE_SIZE (preferred)
        - core >= 1 when large overlap leaves less than MIN_TILE_SIZE
        - core + 2 * buffer_pixels <= resource-derived maximum
    """
    max_full = compute_max_tile_side(context=context)
    warning = None
    core = tile_size

    # Enforce maximum: full tile (core + 2*buffer) must fit resource limit
    max_core = max_full - 2 * buffer_pixels
    if max_core < 1:
        warning = f"Buffer {buffer_pixels}px too large for resource limit ({max_full}px). Using minimum feasible core=1"
        return 1, warning

    # Enforce minimum core size (prefer MIN_TILE_SIZE, but allow smaller when overlap is large)
    min_core = MIN_TILE_SIZE if max_core >= MIN_TILE_SIZE else 1
    if core < min_core:
        warning = f"Tile core size {tile_size} below minimum, using {min_core}"
        core = min_core

    if core > max_core:
        core = max_core
        warning = (
            f"Tile core {tile_size} + 2x{buffer_pixels}px buffer exceeds resource limit "
            f"({max_full}px). Using core={core}"
        )

    return core, warning


def generate_tiles(
    rows: int,
    cols: int,
    tile_size: int,
    overlap: int,
) -> list[TileSpec]:
    """
    Generate tile specifications with overlaps for tiled processing.

    Args:
        rows: Total number of rows in raster.
        cols: Total number of columns in raster.
        tile_size: Core tile size in pixels (without overlap).
        overlap: Overlap size in pixels.

    Returns:
        List of TileSpec objects covering the entire raster.
    """
    tiles = []
    n_tiles_row = int(np.ceil(rows / tile_size))
    n_tiles_col = int(np.ceil(cols / tile_size))

    for i in range(n_tiles_row):
        for j in range(n_tiles_col):
            # Core tile bounds
            row_start = i * tile_size
            row_end = min((i + 1) * tile_size, rows)
            col_start = j * tile_size
            col_end = min((j + 1) * tile_size, cols)

            # Calculate overlaps (bounded by raster edges)
            overlap_top = overlap if i > 0 else 0
            overlap_bottom = overlap if row_end < rows else 0
            overlap_left = overlap if j > 0 else 0
            overlap_right = overlap if col_end < cols else 0

            # Full tile bounds with overlap
            row_start_full = max(0, row_start - overlap_top)
            row_end_full = min(rows, row_end + overlap_bottom)
            col_start_full = max(0, col_start - overlap_left)
            col_end_full = min(cols, col_end + overlap_right)

            tiles.append(
                TileSpec(
                    row_start=row_start,
                    row_end=row_end,
                    col_start=col_start,
                    col_end=col_end,
                    row_start_full=row_start_full,
                    row_end_full=row_end_full,
                    col_start_full=col_start_full,
                    col_end_full=col_end_full,
                    overlap_top=overlap_top,
                    overlap_bottom=overlap_bottom,
                    overlap_left=overlap_left,
                    overlap_right=overlap_right,
                )
            )

    return tiles


def _calculate_tiled(
    surface: SurfaceData,
    location: Location,
    weather: Weather,
    human: HumanParams | None = None,
    precomputed: PrecomputedData | None = None,
    tile_size: int = 1024,
    use_anisotropic_sky: bool | None = None,
    conifer: bool = False,
    physics: SimpleNamespace | None = None,
    materials: SimpleNamespace | None = None,
    max_shadow_distance_m: float = MAX_BUFFER_M,
    tile_workers: int | None = None,
    tile_queue_depth: int | None = None,
    prefetch_tiles: bool | None = None,
    progress_callback: Callable[..., Any] | None = None,
) -> SolweigResult:
    """
    Calculate mean radiant temperature using tiled processing for large rasters.

    Processes the raster in tiles with overlapping buffers to ensure accurate
    shadow calculations at tile boundaries.

    Args:
        surface: Surface/terrain data (DSM required).
        location: Geographic location (lat, lon, UTC offset).
        weather: Weather data for a single timestep.
        human: Human body parameters. Uses defaults if not provided.
        precomputed: Optional pre-computed SVF/walls/shadow matrices.
        tile_size: Core tile size in pixels (default 1024).
        use_anisotropic_sky: Use anisotropic sky model.
            If None, follows calculate() default behavior.
        conifer: Treat vegetation as evergreen conifers. Default False.
        physics: Physics parameters. If None, uses bundled defaults.
        materials: Material properties. If None, uses bundled defaults.
        max_shadow_distance_m: Upper bound on shadow reach in meters (default 1000.0).
            The actual buffer is computed from the tallest DSM pixel via
            calculate_buffer_distance(), capped at this value.
        tile_workers: Number of worker threads for tile execution. If None,
            uses adaptive default based on CPU count.
        tile_queue_depth: Extra queued tile tasks beyond active workers.
            If None, defaults to one queue slot per worker when prefetching.
        prefetch_tiles: Whether to prefetch queued tile tasks. If None,
            chooses automatically based on estimated memory pressure.
        progress_callback: Optional callback(tile_idx, total_tiles).

    Returns:
        SolweigResult with Tmrt grid. State is not returned for single-timestep
        tiled mode.
    """

    if human is None:
        human = HumanParams()

    if surface.svf is None and (precomputed is None or precomputed.svf is None):
        raise MissingPrecomputedData(
            "Sky View Factor (SVF) data is required but not available.",
            "Call surface.compute_svf() before calculate(), or use SurfaceData.prepare() "
            "which computes SVF automatically.",
        )

    if use_anisotropic_sky:
        has_shadow_matrices = (precomputed is not None and precomputed.shadow_matrices is not None) or (
            surface.shadow_matrices is not None
        )
        if not has_shadow_matrices:
            raise MissingPrecomputedData(
                "shadow_matrices required for anisotropic sky model",
                "Either set use_anisotropic_sky=False, or provide shadow matrices via "
                "precomputed=PrecomputedData(shadow_matrices=...) or surface.shadow_matrices",
            )

    # Compute derived weather values
    if not weather._derived_computed:
        weather.compute_derived(location)

    rows, cols = surface.shape
    pixel_size = surface.pixel_size

    # Height-aware buffer: use relative max building height (not absolute elevation)
    max_height = surface.max_height
    buffer_m = calculate_buffer_distance(max_height, max_shadow_distance_m=max_shadow_distance_m)
    buffer_pixels = int(np.ceil(buffer_m / pixel_size))
    logger.info(f"Buffer: {buffer_m:.0f}m ({buffer_pixels}px) from max height {max_height:.1f}m")

    # Validate and adjust tile size
    adjusted_tile_size, warning = validate_tile_size(tile_size, buffer_pixels, pixel_size)
    if warning:
        logger.warning(warning)

    # Check if tiling is actually needed
    if rows <= adjusted_tile_size and cols <= adjusted_tile_size:
        logger.info(f"Raster {rows}x{cols} fits in single tile, using non-tiled calculation")
        from .api import _calculate_single

        return _calculate_single(
            surface=surface,
            location=location,
            weather=weather,
            human=human,
            precomputed=precomputed,
            use_anisotropic_sky=use_anisotropic_sky,
            conifer=conifer,
            physics=physics,
            materials=materials,
            max_shadow_distance_m=max_shadow_distance_m,
        )

    # Generate tiles
    tiles = generate_tiles(rows, cols, adjusted_tile_size, buffer_pixels)
    n_tiles = len(tiles)

    from .api import _calculate_single

    logger.info(
        f"Tiled processing: {rows}x{cols} raster, {n_tiles} tiles, "
        f"tile_size={adjusted_tile_size}, buffer={buffer_m:.0f}m ({buffer_pixels}px) from max height {max_height:.1f}m"
    )

    # Initialize output arrays
    tmrt_out = np.full((rows, cols), np.nan, dtype=np.float32)
    shadow_out = np.full((rows, cols), np.nan, dtype=np.float32)
    kdown_out = np.full((rows, cols), np.nan, dtype=np.float32)
    kup_out = np.full((rows, cols), np.nan, dtype=np.float32)
    ldown_out = np.full((rows, cols), np.nan, dtype=np.float32)
    lup_out = np.full((rows, cols), np.nan, dtype=np.float32)

    # Set up progress reporting
    from .progress import ProgressReporter

    _progress = None if progress_callback is not None else ProgressReporter(total=n_tiles, desc="SOLWEIG tiled")

    # Submit tiles in parallel — Rust releases the GIL during compute_timestep.
    from concurrent.futures import ThreadPoolExecutor, as_completed

    n_workers = _resolve_tile_workers(tile_workers, n_tiles)
    effective_prefetch = (
        prefetch_tiles
        if prefetch_tiles is not None
        else _resolve_prefetch_default(n_workers, n_tiles, adjusted_tile_size, buffer_pixels)
    )
    inflight_limit = _resolve_inflight_limit(n_workers, n_tiles, tile_queue_depth, effective_prefetch)
    logger.info(f"Tiled runtime: workers={n_workers}, inflight_limit={inflight_limit}, prefetch={effective_prefetch}")
    completed = 0

    with ThreadPoolExecutor(max_workers=n_workers) as executor:
        futures: dict[Any, tuple[int, TileSpec]] = {}
        submit_times: dict[Any, float] = {}
        max_queue = 0
        turnaround_sum = 0.0

        def _submit_tile(tile_idx: int, tile: TileSpec) -> None:
            tile_surface = _extract_tile_surface(surface, tile, pixel_size, precomputed=precomputed)
            tile_precomputed = _slice_tile_precomputed(precomputed, tile)

            future = executor.submit(
                _calculate_single,
                surface=tile_surface,
                location=location,
                weather=weather,
                human=human,
                precomputed=tile_precomputed,
                use_anisotropic_sky=use_anisotropic_sky,
                conifer=conifer,
                state=None,
                physics=physics,
                materials=materials,
                max_shadow_distance_m=max_shadow_distance_m,
                return_state_copy=False,
            )
            futures[future] = (tile_idx, tile)
            submit_times[future] = time.perf_counter()

        next_tile = 0
        while next_tile < n_tiles and len(futures) < inflight_limit:
            tile = tiles[next_tile]
            _submit_tile(next_tile, tile)
            next_tile += 1

        while futures:
            future = next(as_completed(futures))
            tile_idx, tile = futures.pop(future)
            submit_t = submit_times.pop(future)
            tile_result = future.result()
            _write_tile_result(tile_result, tile, tmrt_out, shadow_out, kdown_out, kup_out, ldown_out, lup_out)

            turnaround_sum += time.perf_counter() - submit_t
            completed += 1
            if _progress is not None:
                _progress.set_text(f"Tile {completed}/{n_tiles}")
                _progress.update(1)
            if progress_callback:
                progress_callback(completed, n_tiles)

            while next_tile < n_tiles and len(futures) < inflight_limit:
                tile = tiles[next_tile]
                _submit_tile(next_tile, tile)
                next_tile += 1
            max_queue = max(max_queue, max(0, len(futures) - n_workers))

    if _progress is not None:
        _progress.close()
    if completed > 0:
        mean_turnaround_ms = (turnaround_sum / completed) * 1000.0
        logger.info(f"Tiled telemetry: mean_turnaround={mean_turnaround_ms:.1f}ms, max_queue={max_queue}")

    return SolweigResult(
        tmrt=tmrt_out,
        shadow=shadow_out,
        kdown=kdown_out,
        kup=kup_out,
        ldown=ldown_out,
        lup=lup_out,
        utci=None,
        pet=None,
        state=None,
    )


def _calculate_timeseries_tiled(
    surface: SurfaceData,
    weather_series: list[Weather],
    location: Location,
    config: ModelConfig | None = None,
    human: HumanParams | None = None,
    precomputed: PrecomputedData | None = None,
    use_anisotropic_sky: bool | None = None,
    conifer: bool = False,
    physics: SimpleNamespace | None = None,
    materials: SimpleNamespace | None = None,
    wall_material: str | None = None,
    max_shadow_distance_m: float | None = None,
    tile_workers: int | None = None,
    tile_queue_depth: int | None = None,
    prefetch_tiles: bool | None = None,
    *,
    output_dir: str | Path,
    outputs: list[str] | None = None,
    heat_thresholds_day: list[float] | None = None,
    heat_thresholds_night: list[float] | None = None,
    progress_callback: Callable[[int, int], None] | None = None,
) -> TimeseriesSummary:
    """
    Calculate Tmrt timeseries using tiled processing for large rasters.

    Automatically divides large rasters into overlapping tiles and processes
    each timestep tile-by-tile, preserving thermal state accumulation across
    both tiles and timesteps.

    This function is called automatically by calculate() when the
    raster exceeds the resource-derived maximum tile side in either dimension.

    Args:
        surface: Surface/terrain data (DSM required).
        weather_series: List of Weather objects in chronological order.
        location: Geographic location (lat, lon, UTC offset).
        config: Model configuration (provides defaults for None params).
        human: Human body parameters. If None, uses config or defaults.
        precomputed: Optional pre-computed SVF/walls/shadow matrices.
        use_anisotropic_sky: Use anisotropic sky model.
            Shadow matrices are spatially sliced per tile.
        conifer: Treat vegetation as evergreen conifers. Default False.
        physics: Physics parameters. If None, uses config or bundled defaults.
        materials: Material properties. If None, uses config or bundled defaults.
        wall_material: Wall material type for temperature model.
        max_shadow_distance_m: Upper bound on shadow reach in meters.
            If None, uses config or default (1000.0). The actual buffer is
            computed from the tallest DSM pixel via calculate_buffer_distance().
        tile_workers: Number of worker threads for tile execution. If None,
            uses config.tile_workers or adaptive default.
        tile_queue_depth: Extra queued tile tasks beyond active workers.
            If None, uses config.tile_queue_depth or runtime default.
        prefetch_tiles: Whether to prefetch queued tile tasks. If None,
            uses config.prefetch_tiles or defaults to True.
        output_dir: Working directory for all output. Summary grids are always
            saved to ``output_dir/summary/``.
        outputs: Which per-timestep outputs to save as GeoTIFFs
            (e.g., ``["tmrt", "shadow"]``). If None, only summary grids are saved.
        heat_thresholds_day: UTCI thresholds (°C) for daytime exceedance hours.
            Default ``[32, 38]``.
        heat_thresholds_night: UTCI thresholds (°C) for nighttime exceedance hours.
            Default ``[26]``.
        progress_callback: Optional callback(current_step, total_steps).

    Returns:
        :class:`TimeseriesSummary` with aggregated grids and metadata.
    """
    if not weather_series:
        return TimeseriesSummary.empty()

    anisotropic_requested_explicitly = use_anisotropic_sky is True

    # Resolve effective parameters from config
    from ._orchestration import resolve_config_params

    _resolved = resolve_config_params(
        config=config,
        use_anisotropic_sky=use_anisotropic_sky,
        human=human,
        physics=physics,
        materials=materials,
        outputs=outputs,
        max_shadow_distance_m=max_shadow_distance_m,
        tile_workers=tile_workers,
        tile_queue_depth=tile_queue_depth,
        prefetch_tiles=prefetch_tiles,
    )
    effective_aniso = _resolved["use_anisotropic_sky"]
    effective_human = _resolved["human"] if _resolved["human"] is not None else HumanParams()
    effective_physics = _resolved["physics"]
    effective_materials = _resolved["materials"]
    effective_outputs = _resolved["outputs"]
    effective_max_shadow = (
        _resolved["max_shadow_distance_m"] if _resolved["max_shadow_distance_m"] is not None else MAX_BUFFER_M
    )
    effective_tile_workers = _resolved["tile_workers"]
    effective_tile_queue_depth = _resolved["tile_queue_depth"]
    effective_prefetch_tiles = _resolved["prefetch_tiles"]
    anisotropic_arg = effective_aniso if (anisotropic_requested_explicitly or effective_aniso is False) else None

    requested_outputs: set[str] = {"tmrt", "shadow"}
    if effective_outputs:
        requested_outputs |= set(effective_outputs)

    need_shadow = "shadow" in requested_outputs
    need_kdown = "kdown" in requested_outputs
    need_kup = "kup" in requested_outputs
    need_ldown = "ldown" in requested_outputs
    need_lup = "lup" in requested_outputs

    # Fill NaN in surface layers
    surface.fill_nan()

    if surface.svf is None and (precomputed is None or precomputed.svf is None):
        raise MissingPrecomputedData(
            "Sky View Factor (SVF) data is required but not available.",
            "Call surface.compute_svf() before calculate(), or use SurfaceData.prepare() "
            "which computes SVF automatically.",
        )

    if anisotropic_requested_explicitly and effective_aniso:
        has_shadow_matrices = (precomputed is not None and precomputed.shadow_matrices is not None) or (
            surface.shadow_matrices is not None
        )
        if not has_shadow_matrices:
            raise MissingPrecomputedData(
                "shadow_matrices required for anisotropic sky model",
                "Either set use_anisotropic_sky=False, or provide shadow matrices via "
                "precomputed=PrecomputedData(shadow_matrices=...) or surface.shadow_matrices",
            )

    rows, cols = surface.shape
    pixel_size = surface.pixel_size

    # Height-aware buffer: use relative max building height (not absolute elevation)
    max_height = surface.max_height
    buffer_m = calculate_buffer_distance(max_height, max_shadow_distance_m=effective_max_shadow)
    buffer_pixels = int(np.ceil(buffer_m / pixel_size))
    logger.info(f"Buffer: {buffer_m:.0f}m ({buffer_pixels}px) from max height {max_height:.1f}m")

    # Determine tile size
    tile_size = _calculate_auto_tile_size(rows, cols)
    adjusted_tile_size, warning = validate_tile_size(tile_size, buffer_pixels, pixel_size)
    if warning:
        logger.warning(warning)

    # Generate tiles
    tiles = generate_tiles(rows, cols, adjusted_tile_size, buffer_pixels)
    n_tiles = len(tiles)
    n_steps = len(weather_series)

    # Large one-tile runs can underutilize CPU/GPU in timeseries mode.
    # Optionally split into several tiles to increase overlap and throughput.
    if n_tiles == 1:
        adjusted_parallel = _maybe_subdivide_single_tile_for_timeseries(
            rows,
            cols,
            adjusted_tile_size,
            buffer_pixels,
            pixel_size,
            effective_tile_workers,
        )
        if adjusted_parallel < adjusted_tile_size:
            adjusted_tile_size = adjusted_parallel
            tiles = generate_tiles(rows, cols, adjusted_tile_size, buffer_pixels)
            n_tiles = len(tiles)

    # Pre-compute weather (sun positions, radiation)
    from .timeseries import _precompute_weather

    logger.info("=" * 60)
    logger.info("Starting SOLWEIG tiled timeseries calculation")
    logger.info(f"  Grid size: {cols}x{rows} pixels")
    logger.info(f"  Timesteps: {n_steps}")
    start_str = weather_series[0].datetime.strftime("%Y-%m-%d %H:%M")
    end_str = weather_series[-1].datetime.strftime("%Y-%m-%d %H:%M")
    logger.info(f"  Period: {start_str} -> {end_str}")
    logger.info(f"  Location: {location.latitude:.2f}N, {location.longitude:.2f}E")
    logger.info(
        f"  Tiles: {n_tiles} (size={adjusted_tile_size}, buffer={buffer_m:.0f}m from max height {max_height:.1f}m)"
    )
    logger.info("=" * 60)

    logger.info("Pre-computing sun positions and radiation splits...")
    precompute_start = time.time()
    _precompute_weather(weather_series, location)
    precompute_time = time.time() - precompute_start
    logger.info(f"  Pre-computed {n_steps} timesteps in {precompute_time:.1f}s")

    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    use_async_output = bool(effective_outputs) and async_output_enabled()
    _writer = AsyncGeoTiffWriter(output_dir=output_path, surface=surface) if use_async_output else None

    from .api import _calculate_single

    n_workers = _resolve_tile_workers(effective_tile_workers, n_tiles)
    if effective_prefetch_tiles is None:
        effective_prefetch_tiles = _resolve_prefetch_default(n_workers, n_tiles, adjusted_tile_size, buffer_pixels)
    inflight_limit = _resolve_inflight_limit(
        n_workers,
        n_tiles,
        effective_tile_queue_depth,
        effective_prefetch_tiles,
    )
    logger.info(
        f"Tiled runtime: workers={n_workers}, inflight_limit={inflight_limit}, prefetch={effective_prefetch_tiles}"
    )

    # Initialize global state
    state = ThermalState.initial(surface.shape)
    if len(weather_series) >= 2:
        dt0 = weather_series[0].datetime
        dt1 = weather_series[1].datetime
        state.timestep_dec = (dt1 - dt0).total_seconds() / 86400.0
        _timestep_hours = (dt1 - dt0).total_seconds() / 3600.0
    else:
        _timestep_hours = 1.0

    # Grid accumulator for summary statistics
    _accumulator = GridAccumulator(
        shape=surface.shape,
        heat_thresholds_day=heat_thresholds_day if heat_thresholds_day is not None else [32.0, 38.0],
        heat_thresholds_night=heat_thresholds_night if heat_thresholds_night is not None else [26.0],
        timestep_hours=_timestep_hours,
    )

    processed_steps = 0
    total_work = n_steps * n_tiles
    start_time = time.time()

    # Set up progress reporting
    from .progress import ProgressReporter

    _progress = (
        None if progress_callback is not None else ProgressReporter(total=total_work, desc="SOLWEIG tiled timeseries")
    )

    # Pre-create tile data once — surfaces and precomputed data don't change
    # between timesteps. This eliminates N_timesteps × N_tiles redundant copies
    # and allows GVF geometry cache + buffer pool to persist across timesteps.
    tile_surfaces = [_extract_tile_surface(surface, tile, pixel_size, precomputed=precomputed) for tile in tiles]
    tile_precomputeds = [_slice_tile_precomputed(precomputed, tile) for tile in tiles]
    tile_states = [_slice_tile_state(state, tile) for tile in tiles]

    from concurrent.futures import FIRST_COMPLETED, ThreadPoolExecutor, wait

    try:
        with ThreadPoolExecutor(max_workers=n_workers) as executor:
            for t_idx, weather in enumerate(weather_series):
                # Initialize output arrays for this timestep
                tmrt_out = np.full((rows, cols), np.nan, dtype=np.float32)
                shadow_out = np.full((rows, cols), np.nan, dtype=np.float32) if need_shadow else None
                kdown_out = np.full((rows, cols), np.nan, dtype=np.float32) if need_kdown else None
                kup_out = np.full((rows, cols), np.nan, dtype=np.float32) if need_kup else None
                ldown_out = np.full((rows, cols), np.nan, dtype=np.float32) if need_ldown else None
                lup_out = np.full((rows, cols), np.nan, dtype=np.float32) if need_lup else None

                # Refresh preallocated tile states from the current global state.
                for tile_idx, tile in enumerate(tiles):
                    _refresh_tile_state(tile_states[tile_idx], state, tile)

                # Submit tiles in parallel and drain by completion order.
                # This avoids head-of-line blocking when one tile is slower.
                futures: dict[Any, int] = {}
                submit_times: dict[int, float] = {}
                next_submit = 0
                completed_tiles = 0
                max_queue = 0
                turnaround_sum = 0.0

                # Keep only a bounded number of tile tasks in flight.
                while completed_tiles < n_tiles:
                    while next_submit < n_tiles and len(futures) < inflight_limit:
                        future = executor.submit(
                            _calculate_single,
                            surface=tile_surfaces[next_submit],
                            location=location,
                            weather=weather,
                            human=effective_human,
                            precomputed=tile_precomputeds[next_submit],
                            use_anisotropic_sky=anisotropic_arg,
                            conifer=conifer,
                            state=tile_states[next_submit],
                            physics=effective_physics,
                            materials=effective_materials,
                            wall_material=wall_material,
                            max_shadow_distance_m=effective_max_shadow,
                            return_state_copy=False,
                            _requested_outputs=requested_outputs,
                        )
                        futures[future] = next_submit
                        submit_times[next_submit] = time.perf_counter()
                        next_submit += 1
                    max_queue = max(max_queue, max(0, len(futures) - n_workers))

                    done, _ = wait(futures, return_when=FIRST_COMPLETED)
                    if not done:
                        continue
                    for future in done:
                        tile_idx = futures.pop(future)
                        tile = tiles[tile_idx]
                        submit_t = submit_times.pop(tile_idx)

                        if _TIMING_ENABLED:
                            _t0 = time.perf_counter()

                        tile_result = future.result()
                        turnaround_sum += time.perf_counter() - submit_t

                        if _TIMING_ENABLED:
                            _t_ffi = time.perf_counter() - _t0
                            _t1 = time.perf_counter()

                        # Write core results to global arrays (non-overlapping write_slice)
                        _write_tile_result(
                            tile_result,
                            tile,
                            tmrt_out,
                            shadow_out,
                            kdown_out,
                            kup_out,
                            ldown_out,
                            lup_out,
                        )

                        # Merge tile state back to global state (non-overlapping write_slice)
                        if tile_result.state is not None:
                            _merge_tile_state(tile_result.state, tile, state)

                        if _TIMING_ENABLED:
                            _t_merge = time.perf_counter() - _t1
                            print(
                                f"[TIMING] tile {tile_idx + 1}/{n_tiles} "
                                f"ffi={_t_ffi * 1000:.1f}ms "
                                f"merge={_t_merge * 1000:.1f}ms",
                                file=sys.stderr,
                            )

                        # Report progress
                        completed_tiles += 1
                        step = t_idx * n_tiles + completed_tiles
                        if progress_callback is not None:
                            progress_callback(step, total_work)
                        elif _progress is not None:
                            _progress.update(1)

                mean_turnaround_ms = (turnaround_sum / n_tiles) * 1000.0 if n_tiles > 0 else 0.0
                logger.debug(
                    f"Tiled timestep telemetry: step={t_idx + 1}/{n_steps}, "
                    f"mean_turnaround={mean_turnaround_ms:.1f}ms, max_queue={max_queue}"
                )

                # Log timestep completion
                elapsed = time.time() - start_time
                rate = (t_idx + 1) / elapsed if elapsed > 0 else 0
                logger.info(f"  Timestep {t_idx + 1}/{n_steps} complete ({rate:.2f} steps/s)")

                # Create result for this timestep
                result = SolweigResult(
                    tmrt=tmrt_out,
                    shadow=shadow_out,
                    kdown=kdown_out,
                    kup=kup_out,
                    ldown=ldown_out,
                    lup=lup_out,
                    utci=None,
                    pet=None,
                    state=None,  # State managed externally
                )

                # Update grid accumulator (before potential array release)
                _accumulator.update(result, weather, compute_utci_fn=compute_utci_grid)

                from ._orchestration import process_timestep_result

                process_timestep_result(
                    result,
                    weather,
                    effective_outputs,
                    effective_human,
                    _writer,
                    output_dir,
                    surface=surface,
                )
                processed_steps += 1
    finally:
        if _progress is not None:
            _progress.close()
        if _writer is not None:
            _writer.close()

    # Finalize summary
    from ._orchestration import finalize_summary

    return finalize_summary(
        _accumulator,
        surface,
        processed_steps=processed_steps,
        start_time=start_time,
        location=location,
        weather_series=weather_series,
        human=effective_human,
        physics=effective_physics,
        materials=effective_materials,
        use_anisotropic_sky=effective_aniso,
        conifer=conifer,
        output_dir=output_dir,
        outputs=effective_outputs,
        label="(tiled)",
    )



from __future__ import annotations

import numpy as np


def daylen(DOY: int, XLAT: float) -> tuple[float, float, float, float]:
    # Calculation of declination of sun (Eqn. 16). Amplitude= +/-23.45
    # deg. Minimum = DOY 355 (DEC 21), maximum = DOY 172.5 (JUN 21/22).
    # Sun angles.  SOC limited for latitudes above polar circles.
    # Calculate daylength, sunrise and sunset (Eqn. 17)

    RAD = np.pi / 180.0

    DEC = -23.45 * np.cos(2.0 * np.pi * (DOY + 10.0) / 365.0)

    SOC = np.tan(RAD * DEC) * np.tan(RAD * XLAT)
    SOC = min(max(SOC, -1.0), 1.0)
    # SOC=alt

    DAYL = 12.0 + 24.0 * np.arcsin(SOC) / np.pi
    SNUP = 12.0 - DAYL / 2.0
    SNDN = 12.0 + DAYL / 2.0

    return DAYL, DEC, SNDN, SNUP



"""
Internal SVF resolution.

Looks up SVF arrays from ``surface.svf`` or ``precomputed.svf`` at
calculation time. Not part of the public API.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..bundles import DirectionalArrays, SvfBundle

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ..api import PrecomputedData, SurfaceData


def resolve_svf(
    surface: SurfaceData,
    precomputed: PrecomputedData | None,
) -> SvfBundle:
    """
    Look up SVF from surface or precomputed data.

    Returns SvfBundle. Raises MissingPrecomputedData if SVF is not
    available from either source.
    """

    # Priority 1: Check surface.svf (from prepare/cache)
    if surface.svf is not None:
        svf_data = surface.svf
        svf = svf_data.svf
        svf_directional = DirectionalArrays(
            north=svf_data.svf_north,
            east=svf_data.svf_east,
            south=svf_data.svf_south,
            west=svf_data.svf_west,
        )
        svf_veg = svf_data.svf_veg
        svf_veg_directional = DirectionalArrays(
            north=svf_data.svf_veg_north,
            east=svf_data.svf_veg_east,
            south=svf_data.svf_veg_south,
            west=svf_data.svf_veg_west,
        )
        svf_aveg = svf_data.svf_aveg
        svf_aveg_directional = DirectionalArrays(
            north=svf_data.svf_aveg_north,
            east=svf_data.svf_aveg_east,
            south=svf_data.svf_aveg_south,
            west=svf_data.svf_aveg_west,
        )
        svfbuveg = svf_data.svfbuveg

    # Priority 2: Check precomputed.svf (legacy)
    elif precomputed is not None and precomputed.svf is not None:
        svf_data = precomputed.svf
        svf = svf_data.svf
        svf_directional = DirectionalArrays(
            north=svf_data.svf_north,
            east=svf_data.svf_east,
            south=svf_data.svf_south,
            west=svf_data.svf_west,
        )
        svf_veg = svf_data.svf_veg
        svf_veg_directional = DirectionalArrays(
            north=svf_data.svf_veg_north,
            east=svf_data.svf_veg_east,
            south=svf_data.svf_veg_south,
            west=svf_data.svf_veg_west,
        )
        svf_aveg = svf_data.svf_aveg
        svf_aveg_directional = DirectionalArrays(
            north=svf_data.svf_aveg_north,
            east=svf_data.svf_aveg_east,
            south=svf_data.svf_aveg_south,
            west=svf_data.svf_aveg_west,
        )
        svfbuveg = svf_data.svfbuveg

    # No SVF available — require explicit computation
    else:
        from ..errors import MissingPrecomputedData

        raise MissingPrecomputedData(
            "Sky View Factor (SVF) data is required but not available.",
            "Call surface.compute_svf() before calculate(), or use SurfaceData.prepare() "
            "which computes SVF automatically.",
        )

    # Compute svfalfa (SVF angle) from SVF values
    # Formula: svfalfa = arcsin(exp(log(1 - (svf + svf_veg - 1)) / 2))
    # Used in anisotropic sky calculations
    tmp = np.clip(svf + svf_veg - 1.0, 0.0, 1.0)
    safe_term = np.clip(1.0 - tmp, 0.0, 1.0)
    svfalfa = np.arcsin(np.sqrt(safe_term))

    # Construct bundle
    bundle = SvfBundle(
        svf=svf,
        svf_directional=svf_directional,
        svf_veg=svf_veg,
        svf_veg_directional=svf_veg_directional,
        svf_aveg=svf_aveg,
        svf_aveg_directional=svf_aveg_directional,
        svfbuveg=svfbuveg,
        svfalfa=svfalfa,
    )

    return bundle


def adjust_svfbuveg_with_psi(
    svf: NDArray[np.floating],
    svf_veg: NDArray[np.floating],
    psi: float,
    use_veg: bool,
) -> NDArray[np.floating]:
    """
    Adjust svfbuveg with vegetation transmissivity.

    This is needed when SVF was computed fresh without knowledge of psi.

    Args:
        svf: Total sky view factor
        svf_veg: Vegetation-only SVF
        psi: Vegetation transmissivity (0.03 typical for deciduous trees)
        use_veg: Whether vegetation is active

    Returns:
        Adjusted svfbuveg array

    Formula:
        svfbuveg = svf - (1 - svf_veg) * (1 - psi)
    """
    if use_veg:
        svfbuveg = svf - (1.0 - svf_veg) * (1.0 - psi)
        return np.clip(svfbuveg, 0.0, 1.0).astype(np.float32)
    else:
        return svf.astype(np.float32)



"""Shared helpers for timeseries and tiled orchestration paths.

These thin wrappers deduplicate the per-timestep result processing
(UTCI/PET computation, file output, array release) and the
post-loop summary finalization (logging, GeoTIFF export, metadata)
that are identical in both ``timeseries._calculate_timeseries`` and
``tiling._calculate_timeseries_tiled``.

Internal module -- not part of the public API.
"""

from __future__ import annotations

import time
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from .metadata import create_run_metadata, save_run_metadata
from .postprocess import compute_pet_grid, compute_utci_grid
from .solweig_logging import get_logger

if TYPE_CHECKING:
    from .models import HumanParams, SolweigResult, SurfaceData, Weather
    from .output_async import AsyncGeoTiffWriter
    from .summary import GridAccumulator, TimeseriesSummary

logger = get_logger(__name__)


def resolve_config_params(
    *,
    config,
    use_anisotropic_sky: bool | None,
    human,
    physics,
    materials,
    outputs: list[str] | None,
    max_shadow_distance_m: float | None,
    tile_workers: int | None,
    tile_queue_depth: int | None,
    prefetch_tiles: bool | None,
) -> dict:
    """Resolve effective parameters: explicit params > config fallback > defaults.

    Returns a dict with keys matching the parameter names, all resolved
    to their effective values. Callers unpack the result.
    """
    effective_aniso = use_anisotropic_sky
    effective_human = human
    effective_physics = physics
    effective_materials = materials
    effective_outputs = outputs
    effective_max_shadow = max_shadow_distance_m
    effective_tile_workers = tile_workers
    effective_tile_queue_depth = tile_queue_depth
    effective_prefetch_tiles = prefetch_tiles

    if config is not None:
        if effective_aniso is None:
            effective_aniso = config.use_anisotropic_sky
        if effective_human is None:
            effective_human = config.human
        if effective_physics is None:
            effective_physics = config.physics
        if effective_materials is None:
            effective_materials = config.materials
        if effective_outputs is None and config.outputs:
            effective_outputs = config.outputs
        if effective_max_shadow is None:
            effective_max_shadow = config.max_shadow_distance_m
        if effective_tile_workers is None:
            effective_tile_workers = config.tile_workers
        if effective_tile_queue_depth is None:
            effective_tile_queue_depth = config.tile_queue_depth
        if effective_prefetch_tiles is None:
            effective_prefetch_tiles = config.prefetch_tiles

    if effective_aniso is None:
        effective_aniso = True
    if effective_physics is None:
        from .loaders import load_physics

        effective_physics = load_physics()
    if effective_materials is None:
        from .loaders import load_params

        effective_materials = load_params()

    return {
        "use_anisotropic_sky": effective_aniso,
        "human": effective_human,
        "physics": effective_physics,
        "materials": effective_materials,
        "outputs": effective_outputs,
        "max_shadow_distance_m": effective_max_shadow,
        "tile_workers": effective_tile_workers,
        "tile_queue_depth": effective_tile_queue_depth,
        "prefetch_tiles": effective_prefetch_tiles,
    }


def process_timestep_result(
    result: SolweigResult,
    weather: Weather,
    outputs: list[str] | None,
    human: HumanParams | None,
    writer: AsyncGeoTiffWriter | None,
    output_dir: str | Path,
    surface: SurfaceData | None = None,
) -> None:
    """Compute optional UTCI/PET, write per-timestep files, then free arrays.

    This is called after the grid accumulator has already been updated.

    Args:
        result: Single-timestep computation result.
        weather: Weather for this timestep.
        outputs: Requested output layer names (may be None).
        human: Human body params (needed for PET).
        writer: Async GeoTIFF writer (None to use sync fallback).
        output_dir: Output directory path.
        surface: Surface data (needed for sync GeoTIFF CRS/transform).
    """
    from .output_async import collect_output_arrays

    # Compute per-timestep UTCI/PET if requested for file output
    if outputs is not None and "utci" in outputs and result.utci is None:
        result.utci = compute_utci_grid(result.tmrt, weather.ta, weather.rh, weather.ws)
    if outputs is not None and "pet" in outputs and result.pet is None:
        result.pet = compute_pet_grid(result.tmrt, weather.ta, weather.rh, weather.ws, human)

    # Save per-timestep outputs to disk
    if writer is not None and outputs:
        writer.submit(
            timestamp=weather.datetime,
            arrays=collect_output_arrays(result, outputs),
        )
    elif outputs:
        result.to_geotiff(
            output_dir=output_dir,
            timestamp=weather.datetime,
            outputs=outputs,
            surface=surface,
        )

    # Free all large arrays after accumulation + disk write
    result.tmrt = None  # type: ignore[assignment]
    result.shadow = None
    result.kdown = None
    result.kup = None
    result.ldown = None
    result.lup = None
    result.utci = None
    result.pet = None


def finalize_summary(
    accumulator: GridAccumulator,
    surface: SurfaceData,
    *,
    processed_steps: int,
    start_time: float,
    location,
    weather_series: list,
    human,
    physics,
    materials,
    use_anisotropic_sky: bool,
    conifer: bool,
    output_dir: str | Path,
    outputs: list[str] | None,
    label: str = "",
) -> TimeseriesSummary:
    """Finalize grid accumulator, log statistics, save GeoTIFFs and metadata.

    Args:
        accumulator: Grid accumulator with per-timestep updates applied.
        surface: Surface data (for GeoTIFF CRS/transform).
        processed_steps: Number of completed timesteps.
        start_time: ``time.time()`` when the loop started.
        location: Location object.
        weather_series: Full weather list (for metadata).
        human, physics, materials: Model parameters (for metadata).
        use_anisotropic_sky: Whether anisotropic sky was used.
        conifer: Whether conifer mode was used.
        output_dir: Output directory path.
        outputs: Requested output layer names.
        label: Extra text appended to completion log line (e.g. "(tiled)").

    Returns:
        Finalized :class:`TimeseriesSummary`.
    """
    summary = accumulator.finalize()
    summary._surface = surface

    total_time = time.time() - start_time
    overall_rate = processed_steps / total_time if total_time > 0 else 0

    logger.info("=" * 60)
    suffix = f" {label}" if label else ""
    logger.info(f"Calculation complete: {processed_steps} timesteps processed{suffix}")
    logger.info(f"  Total time: {total_time:.1f}s ({overall_rate:.2f} steps/s)")
    if summary.n_timesteps > 0:
        _valid_mean = np.nanmean(summary.tmrt_mean)
        _valid_min = np.nanmin(summary.tmrt_min)
        _valid_max = np.nanmax(summary.tmrt_max)
        logger.info(f"  Tmrt range: {_valid_min:.1f}C - {_valid_max:.1f}C (mean: {_valid_mean:.1f}C)")

    if outputs is not None:
        file_count = processed_steps * len(outputs)
        logger.info(f"  Files saved: {file_count} GeoTIFFs in {output_dir}")
    logger.info("=" * 60)

    summary.to_geotiff(output_dir, surface=surface)
    summary._output_dir = Path(output_dir)
    metadata = create_run_metadata(
        surface=surface,
        location=location,
        weather_series=weather_series,
        human=human,
        physics=physics,
        materials=materials,
        use_anisotropic_sky=use_anisotropic_sky,
        conifer=conifer,
        output_dir=output_dir,
        outputs=outputs,
    )
    save_run_metadata(metadata, output_dir)

    return summary



"""
Ground temperature model component.

Implements the SOLWEIG TgMaps ground temperature model with:
- Parameterization from land cover properties
- Diurnal temperature cycle based on sun altitude
- Clearness index correction for cloudy conditions

Reference:
- Lindberg et al. (2008, 2016) - SOLWEIG ground temperature parameterization
- Reindl et al. (1990) - Clearness index approach
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..bundles import GroundBundle
from ..physics.clearnessindex_2013b import clearnessindex_2013b
from ..physics.daylen import daylen
from ..physics.diffusefraction import diffusefraction

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ..api import Location, Weather


def compute_ground_temperature(
    weather: Weather,
    location: Location,
    alb_grid: NDArray[np.floating],
    emis_grid: NDArray[np.floating],
    tgk_grid: NDArray[np.floating],
    tstart_grid: NDArray[np.floating],
    tmaxlst_grid: NDArray[np.floating],
    *,
    tgk_wall: float | None = None,
    tstart_wall: float | None = None,
    tmaxlst_wall: float | None = None,
) -> GroundBundle:
    """
    Compute ground and wall temperature deviations from air temperature.

    Uses the SOLWEIG TgMaps model with land-cover-specific parameterization.
    Temperature amplitude depends on max sun altitude and land cover type.
    Clearness index correction accounts for reduced heating under cloudy skies.

    Args:
        weather: Weather data including temperature, radiation, sun position
        location: Geographic location (latitude, longitude) for sunrise calculation
        alb_grid: Albedo per pixel (0-1) from land cover properties
        emis_grid: Emissivity per pixel (0-1) from land cover properties
        tgk_grid: TgK parameter per pixel (temperature gain coefficient)
        tstart_grid: Tstart parameter per pixel (temperature baseline offset)
        tmaxlst_grid: TmaxLST parameter per pixel (hour of maximum temperature)
        tgk_wall: Optional wall TgK parameter. If None, uses cobblestone default (0.37).
        tstart_wall: Optional wall Tstart parameter. If None, uses cobblestone default (-3.41).
        tmaxlst_wall: Optional wall TmaxLST parameter. If None, uses cobblestone default (15.0).

    Returns:
        GroundBundle containing:
            - tg: Ground temperature deviation from air temperature (K)
            - tg_wall: Wall temperature deviation from air temperature (K)
            - ci_tg: Clearness index correction factor (0-1)
            - alb_grid: Albedo grid (passed through for convenience)
            - emis_grid: Emissivity grid (passed through for convenience)

    Reference:
        Lindberg et al. (2008): "Urban Multi-scale Environmental Predictor (UMEP)"
        Formula: Tgamp = TgK * altmax + Tstart
                 Tg = Tgamp * sin(phase * pi/2) * CI_TgG
    """
    from ..errors import InvalidSurfaceData, WeatherDataError
    from ..rustalgos import ground as ground_rust

    # Validate inputs to prevent raw AttributeErrors
    for attr in ("datetime", "ta", "rh", "global_rad", "sun_altitude", "altmax"):
        if not hasattr(weather, attr):
            raise WeatherDataError(
                attr,
                "<missing>",
                f"Weather object is missing required attribute '{attr}'. Ensure compute_derived() has been called.",
            )

    for attr in ("latitude", "longitude", "altitude"):
        if not hasattr(location, attr):
            raise InvalidSurfaceData(
                f"Location object is missing required attribute '{attr}'.",
                field=attr,
            )

    # Day of year and sunrise time
    jday = weather.datetime.timetuple().tm_yday
    _, _, _, snup = daylen(jday, location.latitude)

    # Maximum sun altitude for the day (computed in Weather.compute_derived())
    altmax = weather.altmax

    # Decimal time (fraction of day)
    dectime = (weather.datetime.hour + weather.datetime.minute / 60.0) / 24.0

    # CI_TgG correction for non-clear conditions (Lindberg et al. 2008, Reindl et al. 1990)
    # This accounts for reduced ground heating under cloudy skies
    # Full formula from solweig.py: CI_TgG = (radG / radG0) + (1 - corr)
    zen = (90.0 - weather.sun_altitude) * (np.pi / 180.0)  # zenith in radians
    deg2rad = np.pi / 180.0

    # Get clear sky radiation (I0) from clearnessindex function
    location_dict = {"latitude": location.latitude, "longitude": location.longitude, "altitude": location.altitude}
    i0, _, _, _, _ = clearnessindex_2013b(
        zen, jday, weather.ta, weather.rh / 100.0, weather.global_rad, location_dict, -999.0
    )

    # Calculate clear sky direct and diffuse components
    if i0 > 0 and weather.sun_altitude > 0:
        rad_i0, rad_d0 = diffusefraction(i0, weather.sun_altitude, 1.0, weather.ta, weather.rh)
        # Clear sky global horizontal radiation
        rad_g0 = rad_i0 * np.sin(weather.sun_altitude * deg2rad) + rad_d0
    else:
        rad_g0 = 0.0

    # Zenith angle in degrees
    zen_deg = 90.0 - weather.sun_altitude

    # Call Rust implementation (positional args to match Rust signature)
    tg, tg_wall, ci_tg, alb_grid_out, emis_grid_out = ground_rust.compute_ground_temperature(
        weather.ta,
        weather.sun_altitude,
        altmax,
        dectime,
        snup,
        weather.global_rad,
        rad_g0,
        zen_deg,
        alb_grid.astype(np.float32),
        emis_grid.astype(np.float32),
        tgk_grid.astype(np.float32),
        tstart_grid.astype(np.float32),
        tmaxlst_grid.astype(np.float32),
        tgk_wall=tgk_wall,
        tstart_wall=tstart_wall,
        tmaxlst_wall=tmaxlst_wall,
    )

    return GroundBundle(
        tg=tg,
        tg_wall=float(tg_wall),
        ci_tg=float(ci_tg),
        alb_grid=alb_grid_out,
        emis_grid=emis_grid_out,
    )



"""
Pure numpy implementations of morphological operations — **reference implementation only**.

Not called by the production ``calculate()`` API. The fused Rust pipeline
uses ``crate::morphology`` internally.

Retained for readability, tests, and validation against UMEP.
Originally replaced scipy.ndimage functions to eliminate the scipy dependency,
making the package lighter for QGIS plugin distribution.
"""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray


def rotate_array(
    array: NDArray[np.floating],
    angle: float,
    order: int = 1,
    reshape: bool = False,
    mode: str = "nearest",
) -> NDArray[np.floating]:
    """
    Rotate a 2D array by the given angle (in degrees).

    Pure numpy implementation replacing scipy.ndimage.interpolation.rotate.

    Args:
        array: 2D input array to rotate.
        angle: Rotation angle in degrees (counter-clockwise).
        order: Interpolation order (0=nearest, 1=bilinear).
        reshape: If True, output shape is adjusted to contain the whole rotated array.
                 If False (default), output has same shape as input.
        mode: How to handle boundaries ('nearest', 'constant').

    Returns:
        Rotated array.
    """
    if reshape:
        raise NotImplementedError("reshape=True not implemented")

    rows, cols = array.shape
    # scipy uses pixel-centered coordinates: center is at (n-1)/2 for n pixels
    center_y, center_x = (rows - 1) / 2, (cols - 1) / 2

    # Convert angle to radians
    theta = np.radians(angle)
    cos_t, sin_t = np.cos(theta), np.sin(theta)

    # Create output array
    output = np.zeros_like(array)

    # Create coordinate grids
    y_indices, x_indices = np.mgrid[0:rows, 0:cols]

    # Translate to center, rotate, translate back (inverse mapping)
    # For each output pixel, find the corresponding input pixel
    x_centered = x_indices - center_x
    y_centered = y_indices - center_y

    # Inverse rotation to find source coordinates
    # scipy.ndimage.rotate uses counter-clockwise in image coordinates (y pointing down)
    # For inverse mapping, we apply the transpose of the rotation matrix
    src_x = cos_t * x_centered - sin_t * y_centered + center_x
    src_y = sin_t * x_centered + cos_t * y_centered + center_y

    if order == 0:
        # Nearest neighbor interpolation
        src_x_int = np.round(src_x).astype(np.int32)
        src_y_int = np.round(src_y).astype(np.int32)

        # Clip to valid range
        src_x_int = np.clip(src_x_int, 0, cols - 1)
        src_y_int = np.clip(src_y_int, 0, rows - 1)

        output = array[src_y_int, src_x_int]

    elif order == 1:
        # Bilinear interpolation
        x0 = np.floor(src_x).astype(np.int32)
        y0 = np.floor(src_y).astype(np.int32)
        x1 = x0 + 1
        y1 = y0 + 1

        # Clip coordinates
        x0_clipped = np.clip(x0, 0, cols - 1)
        x1_clipped = np.clip(x1, 0, cols - 1)
        y0_clipped = np.clip(y0, 0, rows - 1)
        y1_clipped = np.clip(y1, 0, rows - 1)

        # Weights
        wx = src_x - x0
        wy = src_y - y0
        wx = np.clip(wx, 0, 1)
        wy = np.clip(wy, 0, 1)

        # Bilinear interpolation
        output = (
            array[y0_clipped, x0_clipped] * (1 - wx) * (1 - wy)
            + array[y0_clipped, x1_clipped] * wx * (1 - wy)
            + array[y1_clipped, x0_clipped] * (1 - wx) * wy
            + array[y1_clipped, x1_clipped] * wx * wy
        )
    else:
        raise ValueError(f"order must be 0 or 1, got {order}")

    return output.astype(array.dtype)


def binary_dilation(
    input_array: NDArray[np.bool_],
    structure: NDArray[np.bool_] | None = None,
    iterations: int = 1,
) -> NDArray[np.bool_]:
    """
    Perform binary dilation on a 2D boolean array.

    Pure numpy implementation replacing scipy.ndimage.binary_dilation.

    Args:
        input_array: 2D boolean array to dilate.
        structure: Structuring element (3x3 boolean array).
                   If None, uses 8-connectivity (all neighbors).
        iterations: Number of times to apply dilation.

    Returns:
        Dilated boolean array.
    """
    if structure is None:
        # Default: 8-connectivity (3x3 all ones)
        structure = np.ones((3, 3), dtype=bool)

    result = input_array.copy()

    for _ in range(iterations):
        # Pad the array
        padded = np.pad(result, 1, mode="constant", constant_values=False)
        new_result = np.zeros_like(result)

        # Apply structuring element
        rows, cols = result.shape
        struct_rows, struct_cols = structure.shape
        offset_r = struct_rows // 2
        offset_c = struct_cols // 2

        for dr in range(struct_rows):
            for dc in range(struct_cols):
                if structure[dr, dc]:
                    shifted = padded[
                        1 + dr - offset_r : 1 + rows + dr - offset_r,
                        1 + dc - offset_c : 1 + cols + dc - offset_c,
                    ]
                    new_result |= shifted

        result = new_result

    return result


def generate_binary_structure(rank: int, connectivity: int) -> NDArray[np.bool_]:
    """
    Generate a binary structuring element for morphological operations.

    Pure numpy implementation replacing scipy.ndimage.generate_binary_structure.

    Args:
        rank: Number of dimensions (must be 2).
        connectivity: 1 for 4-connectivity (cross), 2 for 8-connectivity (square).

    Returns:
        3x3 boolean structuring element.
    """
    if rank != 2:
        raise ValueError(f"Only rank=2 supported, got {rank}")

    if connectivity == 1:
        # 4-connectivity (cross pattern)
        return np.array(
            [
                [False, True, False],
                [True, True, True],
                [False, True, False],
            ],
            dtype=bool,
        )
    elif connectivity == 2:
        # 8-connectivity (all neighbors)
        return np.ones((3, 3), dtype=bool)
    else:
        raise ValueError(f"connectivity must be 1 or 2, got {connectivity}")



"""
SOLWEIG computation components.

The production pipeline uses the fused Rust path (``pipeline.compute_timestep``),
which performs shadows, ground temp, GVF, radiation, and Tmrt in a single FFI call.

Active Python modules still called by the fused path:

- **svf_resolution** — ``resolve_svf()``, ``adjust_svfbuveg_with_psi()``
- **shadows** — ``compute_transmissivity()``
- **gvf** — ``detect_building_mask()``
- **ground** — ``compute_ground_temperature()`` (also used by tests)
"""

__all__ = ["ground", "svf_resolution", "shadows", "gvf"]



"""
Sky vault patch geometry — **reference implementation only**.

Not called by the production ``calculate()`` API. The fused Rust pipeline
constructs patch geometry internally.

Retained for readability, tests, and validation against UMEP.
"""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray


def create_patches(
    patch_option: int,
) -> tuple[
    NDArray[np.floating],
    NDArray[np.floating],
    NDArray[np.integer],
    NDArray[np.integer],
    NDArray[np.integer],
    NDArray[np.floating],
    NDArray[np.integer],
]:
    # patch_option = 1 = 145 patches (Robinson & Stone, 2004)
    # patch_option = 2 = 153 patches (Wallenberg et al., 2022)
    # patch_option = 3 = 305 patches -> test
    # patch_option = 4 = 609 patches -> test

    skyvaultalt = np.atleast_2d([])
    skyvaultazi = np.atleast_2d([])

    # Creating skyvault of patches of constant radians (Tregeneza and Sharples, 1993)
    # Patch option 1, 145 patches, Original Robinson & Stone (2004) after Tregenza (1987)/Tregenza & Sharples (1993)
    if patch_option == 1:
        annulino = np.array([0, 12, 24, 36, 48, 60, 72, 84, 90])
        skyvaultaltint = np.array([6, 18, 30, 42, 54, 66, 78, 90])  # Robinson & Stone (2004)
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0])  # Fredrik/Nils
        patches_in_band = np.array([30, 30, 24, 24, 18, 12, 6, 1])  # Robinson & Stone (2004)
    # Patch option 2, 153 patches, Wallenberg et al. (2022)
    elif patch_option == 2:
        annulino = np.array([0, 12, 24, 36, 48, 60, 72, 84, 90])
        skyvaultaltint = np.array([6, 18, 30, 42, 54, 66, 78, 90])  # Robinson & Stone (2004)
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0])  # Fredrik/Nils
        patches_in_band = np.array([31, 30, 28, 24, 19, 13, 7, 1])  # Nils
    # Patch option 3, 305 patches, test
    elif patch_option == 3:
        annulino = np.array([0, 12, 24, 36, 48, 60, 72, 84, 90])
        skyvaultaltint = np.array([6, 18, 30, 42, 54, 66, 78, 90])  # Robinson & Stone (2004)
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0])  # Fredrik/Nils
        patches_in_band = np.array([31 * 2, 30 * 2, 28 * 2, 24 * 2, 19 * 2, 13 * 2, 7 * 2, 1])  # Nils
    # Patch option 4, 609 patches, test
    elif patch_option == 4:
        annulino = np.array([0, 4.5, 9, 15, 21, 27, 33, 39, 45, 51, 57, 63, 69, 75, 81, 90])  # Nils
        skyvaultaltint = np.array([3, 9, 15, 21, 27, 33, 39, 45, 51, 57, 63, 69, 75, 81, 90])  # Nils
        patches_in_band = np.array(
            [
                31 * 2,
                31 * 2,
                30 * 2,
                30 * 2,
                28 * 2,
                28 * 2,
                24 * 2,
                24 * 2,
                19 * 2,
                19 * 2,
                13 * 2,
                13 * 2,
                7 * 2,
                7 * 2,
                1,
            ]
        )  # Nils
        azistart = np.array([0, 0, 4, 4, 2, 2, 5, 5, 8, 8, 0, 0, 10, 10, 0])  # Nils

    skyvaultaziint = np.array([360 / patches for patches in patches_in_band])

    for j in range(0, skyvaultaltint.shape[0]):
        for k in range(0, patches_in_band[j]):
            skyvaultalt = np.append(skyvaultalt, skyvaultaltint[j])
            skyvaultazi = np.append(skyvaultazi, k * skyvaultaziint[j] + azistart[j])

    # skyvaultzen = (90 - skyvaultalt) * deg2rad
    # skyvaultalt = skyvaultalt * deg2rad
    # skyvaultazi = skyvaultazi * deg2rad

    return skyvaultalt, skyvaultazi, annulino, skyvaultaltint, patches_in_band, skyvaultaziint, azistart
